from os import path

import numpy as np
from matplotlib import cm

from wisdem.moorpy.helpers import LineError, CatenaryError, makeTower, rotationMatrix, read_mooring_file
from wisdem.moorpy.Catenary import catenary


class Line:
    """A class for any mooring line that consists of a single material"""

    def __init__(self, mooringSys, num, L, lineType, nSegs=100, cb=0, isRod=0, attachments=[0, 0]):
        """Initialize Line attributes

        Parameters
        ----------
        mooringSys : system object
            The system object that contains the point object
        num : int
            indentifier number
        L : float
            line unstretched length [m]
        lineType : dict
            dictionary containing the coefficients needed to describe the line (could reference an entry of System.lineTypes).
        nSegs : int, optional
            number of segments to split the line into. Used in MoorPy just for plotting. The default is 100.
        cb : float, optional
            line seabed friction coefficient (will be set negative if line is fully suspended). The default is 0.
        isRod : boolean, optional
            determines whether the line is a rod or not. The default is 0.
        attachments : TYPE, optional
            ID numbers of any Points attached to the Line. The default is [0,0]. << consider removing

        Returns
        -------
        None.

        """

        self.sys = mooringSys  # store a reference to the overall mooring system (instance of System class)

        self.number = num
        self.isRod = isRod

        self.L = L  # line unstretched length
        self.type = lineType  # dictionary of a System.lineTypes entry

        self.nNodes = int(nSegs) + 1
        self.cb = float(cb)  # friction coefficient (will automatically be set negative if line is fully suspended)

        self.rA = np.zeros(3)  # end coordinates
        self.rB = np.zeros(3)
        self.fA = np.zeros(3)  # end forces
        self.fB = np.zeros(3)

        # Perhaps this could be made less intrusive by defining it using a line.addpoint() method instead, similar to point.attachline().
        self.attached = (
            attachments  # ID numbers of the Points at the Line ends [a,b] >>> NOTE: not fully supported <<<<
        )
        self.th = 0  # heading of line from end A to B
        self.HF = 0  # fairlead horizontal force saved for next solve
        self.VF = 0  # fairlead vertical force saved for next solve
        self.KA = []  # to be filled with the 2x2 end stiffness matrix from catenary
        self.KB = []  # to be filled with the 2x2 end stiffness matrix from catenary
        self.info = {}  # to hold all info provided by catenary

        self.qs = 1  # flag indicating quasi-static analysis (1). Set to 0 for time series data
        self.show = True  # a flag that will be set to false if we don't want to show the line (e.g. if results missing)
        # print("Created Line "+str(self.number))

    def loadData(self, dirname, rootname, sep=".MD."):
        """Loads line-specific time series data from a MoorDyn output file"""

        self.qs = 0  # signals time series data

        if self.isRod == 1:
            strtype = "Rod"
        elif self.isRod == 0:
            strtype = "Line"

        if path.exists(dirname + rootname + ".MD." + strtype + str(self.number) + ".out"):

            # try:

            # load time series data
            data, ch, channels, units = read_mooring_file(
                dirname + rootname + sep, strtype + str(self.number) + ".out"
            )  # remember number starts on 1 rather than 0

            # get time info
            if "Time" in ch:
                self.Tdata = data[:, ch["Time"]]
                self.dt = self.Tdata[1] - self.Tdata[0]
            else:
                raise LineError("loadData: could not find Time channel for mooring line " + str(self.number))

            nT = len(self.Tdata)  # number of time steps

            # check for position data <<<<<<

            self.xp = np.zeros([nT, self.nNodes])
            self.yp = np.zeros([nT, self.nNodes])
            self.zp = np.zeros([nT, self.nNodes])

            for i in range(self.nNodes):
                self.xp[:, i] = data[:, ch["Node" + str(i) + "px"]]
                self.yp[:, i] = data[:, ch["Node" + str(i) + "py"]]
                self.zp[:, i] = data[:, ch["Node" + str(i) + "pz"]]

            if self.isRod == 0:
                self.Te = np.zeros([nT, self.nNodes - 1])  # read in tension data if available
                if "Seg1Te" in ch:
                    for i in range(self.nNodes - 1):
                        self.Te[:, i] = data[:, ch["Seg" + str(i + 1) + "Te"]]

                self.Ku = np.zeros([nT, self.nNodes])  # read in curvature data if available
                if "Node0Ku" in ch:
                    for i in range(self.nNodes):
                        self.Ku[:, i] = data[:, ch["Node" + str(i) + "Ku"]]

            self.Ux = np.zeros([nT, self.nNodes])  # read in fluid velocity data if available
            self.Uy = np.zeros([nT, self.nNodes])
            self.Uz = np.zeros([nT, self.nNodes])
            if "Node0Ux" in ch:
                for i in range(self.nNodes):
                    self.Ux[:, i] = data[:, ch["Node" + str(i) + "Ux"]]
                    self.Uy[:, i] = data[:, ch["Node" + str(i) + "Uy"]]
                    self.Uz[:, i] = data[:, ch["Node" + str(i) + "Uz"]]

            # Read in tension data if available
            if "Seg1Ten" in ch:
                self.Ten = np.zeros([nT, self.nNodes - 1])
                for i in range(self.nNodes - 1):
                    self.Ten[:, i] = data[:, ch["Seg" + str(i + 1) + "Ten"]]

            self.xpi = self.xp[0, :]
            self.ypi = self.yp[0, :]
            self.zpi = self.zp[0, :]

            # calculate the dynamic LBot !!!!!!! doesn't work for sloped bathymetry yet !!!!!!!!!!
            for i in range(len(self.zp[0])):
                if np.max(self.zp[:, i]) > self.zp[0, 0]:
                    inode = i
                    break
                else:
                    inode = i
            self.LBotDyn = (inode - 1) * self.L / (self.nNodes - 1)

            # get length (constant)
            # self.L = np.sqrt( (self.xpi[-1]-self.xpi[0])**2 + (self.ypi[-1]-self.ypi[0])**2 + (self.zpi[-1]-self.zpi[0])**2 )
            # ^^^^^^^ why are we changing the self.L value to not the unstretched length specified in MoorDyn?
            # moved this below the dynamic LBot calculation because I wanted to use the original self.L
            # >>> this is probably needed for Rods - should look into using for Rods only <<<

            # check for tension data <<<<<<<

            self.show = True

        else:
            self.Tdata = []
            self.show = False

        # >>> this was another option for handling issues - maybe no longer needed <<<
        # except Exception as e:
        #    # don't fail if there's an issue finding data, just flag that the line shouldn't be shown/plotted
        #    print(f"Error geting data for {'Rod' if self.isRod else 'Line'} {self.number}: ")
        #    print(e)
        #    self.show = False

    def getTimestep(self, Time):
        """Get the time step to use for showing time series data"""

        if Time < 0:
            ts = np.int(-Time)  # negative value indicates passing a time step index
        else:  # otherwise it's a time in s, so find closest time step
            for index, item in enumerate(self.Tdata):
                # print "index is "+str(index)+" and item is "+str(item)
                if len(self.Tdata) > 0:
                    ts = -1
                    if item > Time:
                        ts = index
                        break
            if ts == -1:
                raise LineError("getTimestep: requested time likely out of range")

        return ts

    def getLineCoords(self, Time, n=0):  # formerly UpdateLine
        """Gets the updated line coordinates for drawing and plotting purposes."""

        if n == 0:
            n = self.nNodes

        # special temporary case to draw a rod for visualization. This assumes the rod end points have already been set somehow
        if self.qs == 1 and self.isRod > 0:

            # make points for appropriately sized cylinder
            d = self.type["d_vol"]
            Xs, Ys, Zs = makeTower(
                self.L, np.array([d / 2, d / 2])
            )  # add in makeTower method once you start using Rods

            # get unit vector and orientation matrix
            k = (self.rB - self.rA) / self.L
            Rmat = np.array(rotationMatrix(0, np.arctan2(np.hypot(k[0], k[1]), k[2]), np.arctan2(k[1], k[0])))

            # translate and rotate into proper position for Rod
            coords = np.vstack([Xs, Ys, Zs])
            newcoords = np.matmul(Rmat, coords)
            Xs = newcoords[0, :] + self.rA[0]
            Ys = newcoords[1, :] + self.rA[1]
            Zs = newcoords[2, :] + self.rA[2]

            return Xs, Ys, Zs, None

        # if a quasi-static analysis, just call the catenary function to return the line coordinates
        elif self.qs == 1:

            depth = self.sys.depth

            dr = self.rB - self.rA
            LH = np.hypot(dr[0], dr[1])  # horizontal spacing of line ends
            LV = dr[2]  # vertical offset from end A to end B
            if LH > 0:
                cosBeta = dr[0] / LH  # cos of line heading
                sinBeta = dr[1] / LH  # sin of line heading
                self.th = np.arctan2(dr[1], dr[0])  # line heading
            else:  # special case of vertical line: line heading is undefined - use zero as default
                cosBeta = 0.0
                sinBeta = 0.0
                self.th = 0.0

            if np.min([self.rA[2], self.rB[2]]) > -depth:
                self.cb = -depth - np.min(
                    [self.rA[2], self.rB[2]]
                )  # if this line's lower end is off the seabed, set cb negative and to the distance off the seabed
            elif (
                self.cb < 0
            ):  # if a line end is at the seabed, but the cb is still set negative to indicate off the seabed
                self.cb = 0.0  # set to zero so that the line includes seabed interaction.

            try:
                (fAH, fAV, fBH, fBV, info) = catenary(
                    LH,
                    LV,
                    self.L,
                    self.type["EA"],
                    self.type["w"],
                    self.cb,
                    HF0=self.HF,
                    VF0=self.VF,
                    nNodes=n,
                    plots=1,
                )
            except CatenaryError as error:
                raise LineError(self.number, error.message)

            Xs = self.rA[0] + info["X"] * cosBeta
            Ys = self.rA[1] + info["X"] * sinBeta
            Zs = self.rA[2] + info["Z"]
            Ts = info["Te"]
            return Xs, Ys, Zs, Ts

        # otherwise, count on read-in time-series data
        else:

            # figure out what time step to use
            ts = self.getTimestep(Time)

            # drawing rods
            if self.isRod > 0:

                k1 = (
                    np.array(
                        [
                            self.xp[ts, -1] - self.xp[ts, 0],
                            self.yp[ts, -1] - self.yp[ts, 0],
                            self.zp[ts, -1] - self.zp[ts, 0],
                        ]
                    )
                    / self.L
                )  # unit vector

                k = np.array(k1)  # make copy

                Rmat = np.array(
                    rotationMatrix(0, np.arctan2(np.hypot(k[0], k[1]), k[2]), np.arctan2(k[1], k[0]))
                )  # <<< should fix this up at some point, MattLib func may be wrong

                # make points for appropriately sized cylinder
                d = self.type["d_vol"]
                Xs, Ys, Zs = makeTower(
                    self.L, np.array([d / 2, d / 2])
                )  # add in makeTower method once you start using Rods

                # translate and rotate into proper position for Rod
                coords = np.vstack([Xs, Ys, Zs])
                newcoords = np.matmul(Rmat, coords)
                Xs = newcoords[0, :] + self.xp[ts, 0]
                Ys = newcoords[1, :] + self.yp[ts, 0]
                Zs = newcoords[2, :] + self.zp[ts, 0]

                return Xs, Ys, Zs, None

            # drawing lines
            else:

                # handle whether or not there is tension data
                try:  # use average to go from segment tension to node tensions
                    Te = 0.5 * (np.append(self.Te[ts, 0], self.Te[ts, :]) + np.append(self.Te[ts, :], self.Te[ts, -1]))
                except:  # otherwise return zeros to avoid an error (might want a warning in some cases?)
                    Te = np.zeros(self.nNodes)

                return self.xp[ts, :], self.yp[ts, :], self.zp[ts, :], Te

    def getCoordinate(self, s, n=100):
        """Returns position and tension at a specific point along the line's unstretched length"""

        dr = self.rB - self.rA
        LH = np.hypot(dr[0], dr[1])

        Ss = np.linspace(0, self.L, n)
        Xs, Ys, Zs, Ts = self.getLineCoords(0.0, n=n)

        X = np.interp(s, Ss, Xs) * dr[0] / LH
        Y = np.interp(s, Ss, Ys) * dr[1] / LH
        Z = np.interp(s, Ss, Zs)
        T = np.interp(s, Ss, Ts)

        return X, Y, Z, T

    def drawLine2d(
        self,
        Time,
        ax,
        color="k",
        Xuvec=[1, 0, 0],
        Yuvec=[0, 0, 1],
        Xoff=0,
        Yoff=0,
        colortension=False,
        cmap="rainbow",
        plotnodes=[],
        plotnodesline=[],
        label="",
        alpha=1.0,
    ):
        """Draw the line on 2D plot (ax must be 2D)

        Parameters
        ----------
        Time : float
            time value at which to draw the line
        ax : axis
            the axis on which the line is to be drawn
        color : string, optional
            color identifier in one letter (k=black, b=blue,...). The default is "k".
        Xuvec : list, optional
            plane at which the x-axis is desired. The default is [1,0,0].
        Yuvec : list, optional
            plane at which the y-axis is desired. The default is [0,0,1].
        colortension : bool, optional
            toggle to plot the lines in a colormap based on node tensions. The default is False
        cmap : string, optional
            colormap string type to plot tensions when colortension=True. The default is 'rainbow'

        Returns
        -------
        linebit : list
            list of axes and points on which the line can be plotted

        """

        linebit = []  # make empty list to hold plotted lines, however many there are

        if self.isRod > 0:

            Xs, Ys, Zs, Te = self.getLineCoords(Time)

            # apply any 3D to 2D transformation here to provide desired viewing angle
            Xs2d = Xs * Xuvec[0] + Ys * Xuvec[1] + Zs * Xuvec[2]
            Ys2d = Xs * Yuvec[0] + Ys * Yuvec[1] + Zs * Yuvec[2]

            for i in range(int(len(Xs) / 2 - 1)):
                linebit.append(
                    ax.plot(Xs2d[2 * i : 2 * i + 2], Ys2d[2 * i : 2 * i + 2], lw=0.5, color=color)
                )  # side edges
                linebit.append(
                    ax.plot(Xs2d[[2 * i, 2 * i + 2]], Ys2d[[2 * i, 2 * i + 2]], lw=0.5, color=color)
                )  # end A edges
                linebit.append(
                    ax.plot(Xs2d[[2 * i + 1, 2 * i + 3]], Ys2d[[2 * i + 1, 2 * i + 3]], lw=0.5, color=color)
                )  # end B edges

        # drawing lines...
        else:

            if self.qs == 1:
                Xs, Ys, Zs, tensions = self.getLineCoords(Time)
            elif self.qs == 0:
                Xs, Ys, Zs, Ts = self.getLineCoords(Time)
                self.rA = np.array([Xs[0], Ys[0], Zs[0]])
                self.rB = np.array([Xs[-1], Ys[-1], Zs[-1]])
                tensions = self.getLineTens()

            # apply any 3D to 2D transformation here to provide desired viewing angle
            Xs2d = Xs * Xuvec[0] + Ys * Xuvec[1] + Zs * Xuvec[2] + Xoff
            Ys2d = Xs * Yuvec[0] + Ys * Yuvec[1] + Zs * Yuvec[2] + Yoff

            if colortension:  # if the mooring lines want to be plotted with colors based on node tensions
                maxt = np.max(tensions)
                mint = np.min(tensions)
                for i in range(len(Xs) - 1):  # for each node in the line
                    color_ratio = ((tensions[i] + tensions[i + 1]) / 2 - mint) / (
                        maxt - mint
                    )  # ratio of the node tension in relation to the max and min tension
                    cmap_obj = cm.get_cmap(cmap)  # create a cmap object based on the desired colormap
                    rgba = cmap_obj(color_ratio)  # return the rbga values of the colormap of where the node tension is
                    linebit.append(ax.plot(Xs2d[i : i + 2], Ys2d[i : i + 2], color=rgba))
            else:
                linebit.append(
                    ax.plot(Xs2d, Ys2d, lw=1, color=color, label=label, alpha=alpha)
                )  # previously had lw=1 (linewidth)

            if len(plotnodes) > 0:
                for i, node in enumerate(plotnodes):
                    if self.number == plotnodesline[i]:
                        linebit.append(ax.plot(Xs2d[node], Ys2d[node], "o", color=color, markersize=5))

        self.linebit = linebit  # can we store this internally?

        self.X = np.array([Xs, Ys, Zs])

        return linebit

    def drawLine(self, Time, ax, color="k", endpoints=False, shadow=True, colortension=False, cmap_tension="rainbow"):
        """Draw the line in 3D

        Parameters
        ----------
        Time : float
            time value at which to draw the line
        ax : axis
            the axis on which the line is to be drawn
        color : string, optional
            color identifier in one letter (k=black, b=blue,...). The default is "k".
        endpoints : bool, optional
            toggle to plot the end points of the lines. The default is False
        shadow : bool, optional
            toggle to plot the mooring line shadow on the seabed. The default is True
        colortension : bool, optional
            toggle to plot the lines in a colormap based on node tensions. The default is False
        cmap : string, optional
            colormap string type to plot tensions when colortension=True. The default is 'rainbow'

        Returns
        -------
        linebit : list
            list of axes and points on which the line can be plotted
        """

        if not self.show:  # exit if this line isn't set to be shown
            return 0

        if color == "self":
            color = self.color  # attempt to allow custom colors
            lw = self.lw
        else:
            lw = 1

        linebit = []  # make empty list to hold plotted lines, however many there are

        if self.isRod > 0:

            if color == None:
                color = [0.3, 0.3, 0.3]  # if no color provided, default to dark grey rather than rainbow rods

            Xs, Ys, Zs, Ts = self.getLineCoords(Time)

            for i in range(int(len(Xs) / 2 - 1)):
                linebit.append(
                    ax.plot(Xs[2 * i : 2 * i + 2], Ys[2 * i : 2 * i + 2], Zs[2 * i : 2 * i + 2], color=color)
                )  # side edges
                linebit.append(
                    ax.plot(Xs[[2 * i, 2 * i + 2]], Ys[[2 * i, 2 * i + 2]], Zs[[2 * i, 2 * i + 2]], color=color)
                )  # end A edges
                linebit.append(
                    ax.plot(
                        Xs[[2 * i + 1, 2 * i + 3]], Ys[[2 * i + 1, 2 * i + 3]], Zs[[2 * i + 1, 2 * i + 3]], color=color
                    )
                )  # end B edges

            # scatter points for line ends
            # if endpoints == True:
            #    linebit.append(ax.scatter([Xs[0], Xs[-1]], [Ys[0], Ys[-1]], [Zs[0], Zs[-1]], color = color))

        # drawing lines...
        else:

            if self.qs == 1:  # returns the node positions and tensions of the line, doesn't matter what time
                Xs, Ys, Zs, tensions = self.getLineCoords(Time)
            elif self.qs == 0:  # returns the node positions and time data at the given time
                Xs, Ys, Zs, Ts = self.getLineCoords(Time)
                self.rA = np.array([Xs[0], Ys[0], Zs[0]])
                self.rB = np.array([Xs[-1], Ys[-1], Zs[-1]])
                tensions = self.getLineTens()

            if colortension:  # if the mooring lines want to be plotted with colors based on node tensions
                maxt = np.max(tensions)
                mint = np.min(tensions)
                for i in range(len(Xs) - 1):  # for each node in the line
                    color_ratio = ((tensions[i] + tensions[i + 1]) / 2 - mint) / (
                        maxt - mint
                    )  # ratio of the node tension in relation to the max and min tension
                    cmap_obj = cm.get_cmap(cmap_tension)  # create a cmap object based on the desired colormap
                    rgba = cmap_obj(color_ratio)  # return the rbga values of the colormap of where the node tension is
                    linebit.append(ax.plot(Xs[i : i + 2], Ys[i : i + 2], Zs[i : i + 2], color=rgba, zorder=100))
            else:
                linebit.append(ax.plot(Xs, Ys, Zs, color=color, lw=lw, zorder=100))

            if shadow:
                ax.plot(
                    Xs, Ys, np.zeros_like(Xs) - self.sys.depth, color=[0.5, 0.5, 0.5, 0.2], lw=lw, zorder=1.5
                )  # draw shadow

            if endpoints == True:
                linebit.append(ax.scatter([Xs[0], Xs[-1]], [Ys[0], Ys[-1]], [Zs[0], Zs[-1]], color=color))

            # drawing water velocity vectors (not for Rods for now) <<< should handle this better (like in getLineCoords) <<<
            if self.qs == 0:
                ts = self.getTimestep(Time)
                Ux = self.Ux[ts, :]
                Uy = self.Uy[ts, :]
                Uz = self.Uz[ts, :]
                self.Ubits = ax.quiver(Xs, Ys, Zs, Ux, Uy, Uz)  # make quiver plot and save handle to line object

        self.linebit = linebit  # can we store this internally?

        self.X = np.array([Xs, Ys, Zs])

        return linebit

    def redrawLine(self, Time, colortension=False, cmap_tension="rainbow"):  # , linebit):
        """Update 3D line drawing based on instantaneous position"""

        linebit = self.linebit

        if self.isRod > 0:

            Xs, Ys, Zs, Ts = self.getLineCoords(Time)

            for i in range(int(len(Xs) / 2 - 1)):

                linebit[3 * i][0].set_data(
                    Xs[2 * i : 2 * i + 2], Ys[2 * i : 2 * i + 2]
                )  # side edges (x and y coordinates)
                linebit[3 * i][0].set_3d_properties(Zs[2 * i : 2 * i + 2])  #            (z coordinates)
                linebit[3 * i + 1][0].set_data(Xs[[2 * i, 2 * i + 2]], Ys[[2 * i, 2 * i + 2]])  # end A edges
                linebit[3 * i + 1][0].set_3d_properties(Zs[[2 * i, 2 * i + 2]])
                linebit[3 * i + 2][0].set_data(Xs[[2 * i + 1, 2 * i + 3]], Ys[[2 * i + 1, 2 * i + 3]])  # end B edges
                linebit[3 * i + 2][0].set_3d_properties(Zs[[2 * i + 1, 2 * i + 3]])

        # drawing lines...
        else:

            Xs, Ys, Zs, Ts = self.getLineCoords(Time)

            if colortension:
                self.rA = np.array([Xs[0], Ys[0], Zs[0]])  # update the line ends based on the MoorDyn data
                self.rB = np.array([Xs[-1], Ys[-1], Zs[-1]])
                tensions = self.getLineTens()  # get the tensions of the line calculated quasi-statically
                maxt = np.max(tensions)
                mint = np.min(tensions)
                cmap_obj = cm.get_cmap(cmap_tension)  # create the colormap object

                for i in range(
                    len(Xs) - 1
                ):  # for each node in the line, find the relative tension of the segment based on the max and min tensions
                    color_ratio = ((tensions[i] + tensions[i + 1]) / 2 - mint) / (maxt - mint)
                    rgba = cmap_obj(color_ratio)
                    linebit[i][
                        0
                    ]._color = rgba  # set the color of the segment to a new color based on its updated tension
                    linebit[i][0].set_data(Xs[i : i + 2], Ys[i : i + 2])  # set the x and y coordinates
                    linebit[i][0].set_3d_properties(Zs[i : i + 2])  # set the z coorindates

            else:
                linebit[0][0].set_data(Xs, Ys)  # (x and y coordinates)
                linebit[0][0].set_3d_properties(Zs)  # (z coordinates)

            # drawing water velocity vectors (not for Rods for now)
            if self.qs == 0:
                ts = self.getTimestep(Time)
                Ux = self.Ux[ts, :]
                Uy = self.Uy[ts, :]
                Uz = self.Uz[ts, :]
                # segments = quiver_data_to_segments(Xs, Ys, Zs, Ux, Uy, Uz, scale=2)
                # self.Ubits.set_segments(segments)

        return linebit

    def setEndPosition(self, r, endB):
        """Sets the end position of the line based on the input endB value.

        Parameters
        ----------
        r : array
            x,y,z coorindate position vector of the line end [m].
        endB : boolean
            An indicator of whether the r array is at the end or beginning of the line

        Raises
        ------
        LineError
            If the given endB value is not a 1 or 0

        Returns
        -------
        None.

        """

        if endB == 1:
            self.rB = np.array(r, dtype=np.float_)
        elif endB == 0:
            self.rA = np.array(r, dtype=np.float_)
        else:
            raise LineError("setEndPosition: endB value has to be either 1 or 0")

    def staticSolve(self, reset=False, tol=0.0001, profiles=0):
        """Solves static equilibrium of line. Sets the end forces of the line based on the end points' positions.

        Parameters
        ----------
        reset : boolean, optional
            Determines if the previous fairlead force values will be used for the catenary iteration. The default is False.

        tol : float
            Convergence tolerance for catenary solver measured as absolute error of x and z values in m.

        profiles : int
            Values greater than 0 signal for line profile data to be saved (used for plotting, getting distributed tensions, etc).

        Raises
        ------
        LineError
            If the horizontal force at the fairlead (HF) is less than 0

        Returns
        -------
        None.

        """

        depth = self.sys.depth

        dr = self.rB - self.rA
        LH = np.hypot(dr[0], dr[1])  # horizontal spacing of line ends
        LV = dr[2]  # vertical offset from end A to end B
        if LH > 0:
            cosBeta = dr[0] / LH  # cos of line heading
            sinBeta = dr[1] / LH  # sin of line heading
            self.th = np.arctan2(dr[1], dr[0])  # line heading
        else:  # special case of vertical line: line heading is undefined - use zero as default
            cosBeta = 0.0
            sinBeta = 0.0
            self.th = 0.0

        if self.rA[2] < -depth:
            raise LineError("Line {} end A is lower than the seabed.".format(self.number))
        elif self.rB[2] < -depth:
            raise LineError("Line {} end B is lower than the seabed.".format(self.number))
        elif np.min([self.rA[2], self.rB[2]]) > -depth:
            self.cb = -depth - np.min(
                [self.rA[2], self.rB[2]]
            )  # if this line's lower end is off the seabed, set cb negative and to the distance off the seabed
        elif self.cb < 0:  # if a line end is at the seabed, but the cb is still set negative to indicate off the seabed
            self.cb = 0.0  # set to zero so that the line includes seabed interaction.

        if (
            self.HF < 0
        ):  # or self.VF < 0:  <<<<<<<<<<< it shouldn't matter if VF is negative - this could happen for buoyant lines, etc.
            raise LineError("Line HF cannot be negative")  # this could be a ValueError too...

        if reset == True:  # Indicates not to use previous fairlead force values to start catenary
            self.HF = 0  # iteration with, and insteady use the default values.

        try:
            (fAH, fAV, fBH, fBV, info) = catenary(
                LH,
                LV,
                self.L,
                self.type["EA"],
                self.type["w"],
                CB=self.cb,
                Tol=tol,
                HF0=self.HF,
                VF0=self.VF,
                plots=profiles,
            )  # call line model
        except CatenaryError as error:
            raise LineError(self.number, error.message)

        self.HF = info["HF"]
        self.VF = info["VF"]
        self.KA2 = info["stiffnessA"]
        self.KB2 = info["stiffnessB"]
        self.LBot = info["LBot"]
        self.info = info

        self.fA[0] = fAH * cosBeta
        self.fA[1] = fAH * sinBeta
        self.fA[2] = fAV
        self.fB[0] = fBH * cosBeta
        self.fB[1] = fBH * sinBeta
        self.fB[2] = fBV
        self.TA = np.sqrt(fAH * fAH + fAV * fAV)  # end tensions
        self.TB = np.sqrt(fBH * fBH + fBV * fBV)

        # ----- compute 3d stiffness matrix for both line ends (3 DOF + 3 DOF) -----

        # solve for required variables to set up the perpendicular stiffness. Keep it horizontal
        # L_xy = np.linalg.norm(self.rB[:2] - self.rA[:2])
        # T_xy = np.linalg.norm(self.fB[:2])

        # create the rotation matrix based on the heading angle that the line is from the horizontal
        R = rotationMatrix(0, 0, self.th)

        # initialize the line's analytic stiffness matrix in the "in-line" plane then rotate the matrix to be about the global frame [K'] = [R][K][R]^T
        def from2Dto3Drotated(K2D, F, L):
            if L > 0:
                Kt = F / L  # transverse stiffness term
            else:
                Kt = 0.0

            K2 = np.array([[K2D[0, 0], 0, K2D[0, 1]], [0, Kt, 0], [K2D[1, 0], 0, K2D[1, 1]]])
            return np.matmul(np.matmul(R, K2), R.T)

        self.KA = from2Dto3Drotated(
            info["stiffnessA"], -fBH, LH
        )  # stiffness matrix describing reaction force on end A due to motion of end A
        self.KB = from2Dto3Drotated(
            info["stiffnessB"], -fBH, LH
        )  # stiffness matrix describing reaction force on end B due to motion of end B
        self.KAB = from2Dto3Drotated(
            info["stiffnessAB"], fBH, LH
        )  # stiffness matrix describing reaction force on end B due to motion of end A

        # self.K6 = np.block([[ from2Dto3Drotated(self.KA),  from2Dto3Drotated(self.KAB.T)],
        #                    [ from2Dto3Drotated(self.KAB), from2Dto3Drotated(self.KB)  ]])

        if profiles > 1:
            import matplotlib.pyplot as plt

            plt.plot(info["X"], info["Z"])
            plt.show()

    def getEndForce(self, endB):
        """Returns the force of the line at the specified end based on the endB value

        Parameters
        ----------
        endB : boolean
            An indicator of which end of the line is the force wanted

        Raises
        ------
        LineError
            If the given endB value is not a 1 or 0

        Returns
        -------
        fA or fB: array
            The force vector at the end of the line

        """

        if endB == 1:
            return self.fB
        elif endB == 0:
            return self.fA
        else:
            raise LineError("getEndForce: endB value has to be either 1 or 0")

    def getStiffnessMatrix(self):
        """Returns the stiffness matrix of a line derived from analytic terms in the jacobian of catenary

        Raises
        ------
        LineError
            If a singluar matrix error occurs while taking the inverse of the Line's Jacobian matrix.

        Returns
        -------
        K2_rot : matrix
            the analytic stiffness matrix of the line in the rotated frame.

        """

        # take the inverse of the Jacobian to get the starting analytic stiffness matrix
        """
        if np.isnan(self.jacobian[0,0]): #if self.LBot >= self.L and self.HF==0. and self.VF==0.  << handle tricky cases here?
            K = np.array([[0., 0.], [0., 1.0/self.jacobian[1,1] ]])
        else:
            try:
                K = np.linalg.inv(self.jacobian)
            except:
                raise LineError(self.number, f"Check Line Length ({self.L}), it might be too long, or check catenary ProfileType")
        """

        # solve for required variables to set up the perpendicular stiffness. Keep it horizontal
        L_xy = np.linalg.norm(self.rB[:2] - self.rA[:2])
        T_xy = np.linalg.norm(self.fB[:2])
        Kt = T_xy / L_xy

        # initialize the line's analytic stiffness matrix in the "in-line" plane
        KA = np.array([[self.KA2[0, 0], 0, self.KA2[0, 1]], [0, Kt, 0], [self.KA2[1, 0], 0, self.KA2[1, 1]]])

        KB = np.array([[self.KB2[0, 0], 0, self.KB2[0, 1]], [0, Kt, 0], [self.KB2[1, 0], 0, self.KB2[1, 1]]])

        # create the rotation matrix based on the heading angle that the line is from the horizontal
        R = rotationMatrix(0, 0, self.th)

        # rotate the matrix to be about the global frame [K'] = [R][K][R]^T
        KA_rot = np.matmul(np.matmul(R, KA), R.T)
        KB_rot = np.matmul(np.matmul(R, KB), R.T)

        return KA_rot, KB_rot

    def getLineTens(self):
        """Calls the catenary function to return the tensions of the Line for a quasi-static analysis"""

        # >>> this can probably be done using data already generated by static Solve <<<

        depth = self.sys.depth

        dr = self.rB - self.rA
        LH = np.hypot(dr[0], dr[1])  # horizontal spacing of line ends
        LV = dr[2]  # vertical offset from end A to end B

        if np.min([self.rA[2], self.rB[2]]) > -depth:
            self.cb = -depth - np.min(
                [self.rA[2], self.rB[2]]
            )  # if this line's lower end is off the seabed, set cb negative and to the distance off the seabed
        elif self.cb < 0:  # if a line end is at the seabed, but the cb is still set negative to indicate off the seabed
            self.cb = 0.0  # set to zero so that the line includes seabed interaction.

        try:
            (fAH, fAV, fBH, fBV, info) = catenary(
                LH,
                LV,
                self.L,
                self.type["EA"],
                self.type["w"],
                self.cb,
                HF0=self.HF,
                VF0=self.VF,
                nNodes=self.nNodes,
                plots=1,
            )
        except CatenaryError as error:
            raise LineError(self.number, error.message)

        Ts = info["Te"]
        return Ts

    def getTension(self, s):
        """Returns tension at a given point along the line

        Parameters
        ----------

        s : scalar or array-like
            Value or array of values for the arc length along the line from end A to end B at which
            the information is desired. Positive values are arc length in m, negative values are a
            relative location where 0 is end A, -1 is end B, and -0.5 is the midpoint.

        Returns
        -------

        tension value(s)

        """
        # if s < 0:
        #    s = -s*self.L
        # if s > self.L:
        #    raise ValueError('Specified arc length is larger than the line unstretched length.')

        Te = np.interp(s, self.info["s"], self.info["Te"])

        return Te

    def getPosition(self, s):
        """Returns position at a given point along the line

        Parameters
        ----------

        s : scalar or array-like
            Value or array of values for the arc length along the line from end A to end B at which
            the information is desired. Positive values are arc length in m, negative values are a
            relative location where 0 is end A, -1 is end B, and -0.5 is the midpoint.

        Returns
        -------

        position vector(s)

        """

        # >>> should be merged with getLineCoords and getCoordinate functionality <<<

        x = np.interp(s, self.info["s"], self.info["X"])
        z = np.interp(s, self.info["s"], self.info["Z"])

        dr = self.rB - self.rA
        LH = np.hypot(dr[0], dr[1])
        Xs = self.rA[0] + x * dr[0] / LH
        Ys = self.rA[1] + x * dr[1] / LH
        Zs = self.rA[2] + z

        return np.vstack([Xs, Ys, Zs])

    def attachLine(self, lineID, endB):
        pass



#!/usr/bin/env python
# encoding: utf-8
"""
precomp.py

Created by Andrew Ning on 2012-02-28.
Copyright (c)  NREL. All rights reserved.
"""

from __future__ import print_function

import numpy as np
import math
import copy
import os

# from rotorstruc import SectionStrucInterface
# from wisdem.common import sind, cosd
# from external._precomp import precomp as _precomp
from wisdem.rotorse._precomp import precomp as _precomp


def web_loc(r, chord, le, ib_idx, ob_idx, ib_webc, ob_webc):

    n = len(r)
    loc = np.zeros(n)

    for i in range(n):

        if i < ib_idx or i > ob_idx:
            loc[i] = -1
        else:
            xn = (r[i] - r[ib_idx]) / (r[ob_idx] - r[ib_idx])
            loc[i] = (
                le[i]
                - (le[ib_idx] - ib_webc) * chord[ib_idx] / chord[i] * (1 - xn)
                - (le[ob_idx] - ob_webc) * chord[ob_idx] / chord[i] * xn
            )

    return loc


class PreComp:
    def __init__(
        self,
        r,
        chord,
        theta,
        leLoc,
        precurve,
        presweep,
        profile,
        materials,
        upperCS,
        lowerCS,
        websCS,
        sector_idx_strain_spar_ps,
        sector_idx_strain_spar_ss,
        sector_idx_strain_te_ps,
        sector_idx_strain_te_ss,
    ):
        """Constructor

        Parameters
        ----------
        r : ndarray (m)
            radial positions. r[0] should be the hub location
            while r[-1] should be the blade tip. Any number
            of locations can be specified between these in ascending order.
        chord : ndarray (m)
            array of chord lengths at corresponding radial positions
        theta : ndarray (deg)
            array of twist angles at corresponding radial positions.
            (positive twist decreases angle of attack)
        leLoc : ndarray(float)
            array of leading-edge positions from a reference blade axis (usually blade pitch axis).
            locations are normalized by the local chord length.  e.g. leLoc[i] = 0.2 means leading edge
            is 0.2*chord[i] from reference axis.   positive in -x direction for airfoil-aligned coordinate system
        profile : list(:class:`Profile`)
            airfoil shape at each radial position
        materials : list(:class:`Orthotropic2DMaterial`), optional
            list of all Orthotropic2DMaterial objects used in defining the geometry
        upperCS, lowerCS, websCS : list(:class:`CompositeSection`)
            list of CompositeSection objections defining the properties for upper surface, lower surface,
            and shear webs (if any) for each section

        """

        self.r = np.array(r)
        self.chord = np.array(chord)
        self.theta = np.array(theta)
        self.leLoc = np.array(leLoc)
        self.precurve = np.array(precurve)
        self.presweep = np.array(presweep)

        self.profile = profile
        self.materials = materials
        self.upperCS = upperCS
        self.lowerCS = lowerCS
        self.websCS = websCS

        self.sector_idx_strain_spar_ps = sector_idx_strain_spar_ps
        self.sector_idx_strain_spar_ss = sector_idx_strain_spar_ss
        self.sector_idx_strain_te_ps = sector_idx_strain_te_ps
        self.sector_idx_strain_te_ss = sector_idx_strain_te_ss

        # twist rate
        self.th_prime = _precomp.tw_rate(self.r, self.theta)

    def sectionProperties(self):
        """see meth:`SectionStrucInterface.sectionProperties`"""

        # radial discretization
        nsec = len(self.r)

        # initialize variables
        beam_z = self.r
        beam_EA = np.zeros(nsec)
        beam_EIxx = np.zeros(nsec)
        beam_EIyy = np.zeros(nsec)
        beam_EIxy = np.zeros(nsec)
        beam_GJ = np.zeros(nsec)
        beam_rhoA = np.zeros(nsec)
        beam_A = np.zeros(nsec)
        beam_rhoJ = np.zeros(nsec)
        beam_Tw_iner = np.zeros(nsec)

        beam_flap_iner = np.zeros(nsec)
        beam_edge_iner = np.zeros(nsec)

        beam_x_sc = np.zeros(nsec)
        beam_y_sc = np.zeros(nsec)
        beam_x_tc = np.zeros(nsec)
        beam_y_tc = np.zeros(nsec)
        beam_x_cg = np.zeros(nsec)
        beam_y_cg = np.zeros(nsec)

        # distance to elastic center from point about which structural properties are computed
        # using airfoil coordinate system
        beam_x_ec = np.zeros(nsec)
        beam_y_ec = np.zeros(nsec)

        # distance to elastic center from airfoil nose
        # using profile coordinate system
        self.x_ec_nose = np.zeros(nsec)
        self.y_ec_nose = np.zeros(nsec)

        profile = self.profile
        mat = self.materials
        csU = self.upperCS
        csL = self.lowerCS
        csW = self.websCS

        # arrange materials into array
        n = len(mat)
        E1 = [0] * n
        E2 = [0] * n
        G12 = [0] * n
        nu12 = [0] * n
        rho = [0] * n

        for i in range(n):
            E1[i] = mat[i].E1
            E2[i] = mat[i].E2
            G12[i] = mat[i].G12
            nu12[i] = mat[i].nu12
            rho[i] = mat[i].rho

        for i in range(nsec):
            # print(i)

            xnode, ynode = profile[i]._preCompFormat()
            locU, n_laminaU, n_pliesU, tU, thetaU, mat_idxU = csU[i]._preCompFormat()
            locL, n_laminaL, n_pliesL, tL, thetaL, mat_idxL = csL[i]._preCompFormat()
            locW, n_laminaW, n_pliesW, tW, thetaW, mat_idxW = csW[i]._preCompFormat()

            nwebs = len(locW)

            # address a bug in f2py (need to pass in length 1 arrays even though they are not used)
            if nwebs == 0:
                locW = [0]
                n_laminaW = [0]
                n_pliesW = [0]
                tW = [0]
                thetaW = [0]
                mat_idxW = [0]

            results = _precomp.properties(
                self.chord[i],
                self.theta[i],
                self.th_prime[i],
                self.leLoc[i],
                xnode,
                ynode,
                E1,
                E2,
                G12,
                nu12,
                rho,
                locU,
                n_laminaU,
                n_pliesU,
                tU,
                thetaU,
                mat_idxU,
                locL,
                n_laminaL,
                n_pliesL,
                tL,
                thetaL,
                mat_idxL,
                nwebs,
                locW,
                n_laminaW,
                n_pliesW,
                tW,
                thetaW,
                mat_idxW,
            )

            beam_EIxx[i] = results[1]  # EIedge
            beam_EIyy[i] = results[0]  # EIflat
            beam_GJ[i] = results[2]
            beam_EA[i] = results[3]
            beam_EIxy[i] = results[4]  # EIflapedge
            beam_x_sc[i] = results[10]
            beam_y_sc[i] = results[11]
            beam_x_tc[i] = results[12]
            beam_y_tc[i] = results[13]
            beam_x_ec[i] = results[12] - results[10]
            beam_y_ec[i] = results[13] - results[11]
            beam_rhoA[i] = results[14]
            beam_A[i] = results[15]
            beam_rhoJ[i] = results[16] + results[17]  # perpendicular axis theorem
            beam_Tw_iner[i] = results[18]
            beam_x_cg[i] = results[19]
            beam_y_cg[i] = results[20]

            beam_flap_iner[i] = results[16]
            beam_edge_iner[i] = results[17]

            self.x_ec_nose[i] = results[13] + self.leLoc[i] * self.chord[i]
            self.y_ec_nose[i] = results[12]  # switch b.c of coordinate system used

        return (
            beam_EIxx,
            beam_EIyy,
            beam_GJ,
            beam_EA,
            beam_EIxy,
            beam_x_ec,
            beam_y_ec,
            beam_rhoA,
            beam_A,
            beam_rhoJ,
            beam_Tw_iner,
            beam_flap_iner,
            beam_edge_iner,
            beam_x_tc,
            beam_y_tc,
            beam_x_sc,
            beam_y_sc,
            beam_x_cg,
            beam_y_cg,
        )

    def criticalStrainLocations(self, sector_idx_strain_ss, sector_idx_strain_ps):

        n = len(self.r)

        # find location of max thickness on airfoil
        xun = np.zeros(n)
        xln = np.zeros(n)
        yun = np.zeros(n)
        yln = np.zeros(n)

        # for i, p in enumerate(self.profile):
        #     xun[i], yun[i], yln[i] = p.locationOfMaxThickness()
        # xln = xun

        for i in range(n):
            csU = self.upperCS[i]
            csL = self.lowerCS[i]
            pf = self.profile[i]
            idx_ss = sector_idx_strain_ss[i]
            idx_ps = sector_idx_strain_ps[i]

            if idx_ps == None:
                xln[i] = 0.0
                yln[i] = 0.0
            else:
                xln[i] = 0.5 * (csL.loc[idx_ps] + csL.loc[idx_ps + 1])
                yln[i] = np.interp(xln[i], pf.x, pf.yl)

            if idx_ss == None:
                xun[i] = 0.0
                yun[i] = 0.0
            else:
                xun[i] = 0.5 * (csU.loc[idx_ss] + csU.loc[idx_ss + 1])
                yun[i] = np.interp(xun[i], pf.x, pf.yu)

        # make dimensional and define relative to elastic center
        xu = xun * self.chord - self.x_ec_nose
        xl = xln * self.chord - self.x_ec_nose
        yu = yun * self.chord - self.y_ec_nose
        yl = yln * self.chord - self.y_ec_nose

        # switch to airfoil coordinate system
        xu, yu = yu, xu
        xl, yl = yl, xl

        return xu, xl, yu, yl

    def panelBucklingStrain(self, sector_idx_array):
        """
        see chapter on Structural Component Design Techniques from Alastair Johnson
        section 6.2: Design of composite panels

        assumes: large aspect ratio, simply supported, uniaxial compression, flat rectangular plate

        """
        chord = self.chord
        nsec = len(self.r)

        eps_crit = np.zeros(nsec)

        for i in range(nsec):

            cs = self.upperCS[i]  # TODO: lower surface may be the compression one
            sector_idx = sector_idx_array[i]

            if sector_idx == None:
                eps_crit[i] = 0.0

            else:

                # chord-wise length of sector
                sector_length = chord[i] * (cs.loc[sector_idx + 1] - cs.loc[sector_idx])

                # get matrices
                A, B, D, totalHeight = cs.compositeMatrices(sector_idx)
                E = cs.effectiveEAxial(sector_idx)
                D1 = D[0, 0]
                D2 = D[1, 1]
                D3 = D[0, 1] + 2 * D[2, 2]

                # use empirical formula
                Nxx = 2 * (math.pi / sector_length) ** 2 * (math.sqrt(D1 * D2) + D3)
                # Nxx = 3.6 * (math.pi/sector_length)**2 * D1

                eps_crit[i] = -Nxx / totalHeight / E

        return eps_crit


def skipLines(f, n):
    for i in range(n):
        f.readline()


class CompositeSection:
    """A CompositeSection defines the layup of the entire
    airfoil cross-section

    """

    def __init__(self, loc, n_plies, t, theta, mat_idx, materials):
        """Constructor

        Parameters
        ----------


        """

        self.loc = np.array(loc)  # np.array([0.0, 0.15, 0.50, 1.00])

        # should be list of numpy arrays
        self.n_plies = n_plies  # [ [1, 1, 33],  [1, 1, 17, 38, 0, 37, 16], [1, 1, 17, 0, 16] ]
        self.t = t  # [ [0.000381, 0.00051, 0.00053], [0.000381, 0.00051, 0.00053, 0.00053, 0.003125, 0.00053, 0.00053], [0.000381, 0.00051, 0.00053, 0.003125, 0.00053] ]
        self.theta = theta  # [ [0, 0, 20], [0, 0, 20, 30, 0, 30, 20], [0, 0, 20, 0, 0] ]
        self.mat_idx = mat_idx  # [ [3, 4, 2], [3, 4, 2, 1, 5, 1, 2], [3, 4, 2, 5, 2] ]

        self.materials = materials

    def mycopy(self):
        return CompositeSection(
            copy.deepcopy(self.loc),
            copy.deepcopy(self.n_plies),
            copy.deepcopy(self.t),
            copy.deepcopy(self.theta),
            copy.deepcopy(self.mat_idx),
            self.materials,
        )  # TODO: copy materials (for now it never changes so I'm not looking at it)

    @classmethod
    def initFromPreCompLayupFile(cls, fname, locW, materials, readLocW=False):
        """Construct CompositeSection object from a PreComp input file

        Parameters
        ----------
        fname : str
            name of input file
        webLoc : ndarray
            array of web locations (i.e. [0.15, 0.5] has two webs
            one located at 15% chord from the leading edge and
            the second located at 50% chord)
        materials : list(:class:`Orthotropic2DMaterial`)
            material objects defined in same order as used in the input file
            can use :meth:`Orthotropic2DMaterial.initFromPreCompFile`
        readLocW : optionally read web location from main input file rather than
            have the user provide it

        Returns
        -------
        compSec : CompositeSection
            an initialized CompositeSection object

        """

        f = open(fname)

        skipLines(f, 3)

        # number of sectors
        n_sector = int(f.readline().split()[0])

        skipLines(f, 2)

        # read normalized chord locations
        locU = [float(x) for x in f.readline().split()]

        n_pliesU, tU, thetaU, mat_idxU = CompositeSection.__readSectorsFromFile(f, n_sector)
        upper = cls(locU, n_pliesU, tU, thetaU, mat_idxU, materials)

        skipLines(f, 3)

        # number of sectors
        n_sector = int(f.readline().split()[0])

        skipLines(f, 2)

        locL = [float(x) for x in f.readline().split()]

        n_pliesL, tL, thetaL, mat_idxL = CompositeSection.__readSectorsFromFile(f, n_sector)
        lower = cls(locL, n_pliesL, tL, thetaL, mat_idxL, materials)

        skipLines(f, 4)

        if readLocW:
            locW = CompositeSection.__readWebLocFromFile(fname)
        n_sector = len(locW)

        n_pliesW, tW, thetaW, mat_idxW = CompositeSection.__readSectorsFromFile(f, n_sector)
        webs = cls(locW, n_pliesW, tW, thetaW, mat_idxW, materials)

        f.close()

        return upper, lower, webs

    @staticmethod
    def __readSectorsFromFile(f, n_sector):
        """private method"""

        n_plies = [0] * n_sector
        t = [0] * n_sector
        theta = [0] * n_sector
        mat_idx = [0] * n_sector

        for i in range(n_sector):
            skipLines(f, 2)

            line = f.readline()
            if line == "":
                return []  # no webs
            n_lamina = int(line.split()[1])

            skipLines(f, 4)

            n_plies_S = np.zeros(n_lamina)
            t_S = np.zeros(n_lamina)
            theta_S = np.zeros(n_lamina)
            mat_idx_S = np.zeros(n_lamina)

            for j in range(n_lamina):
                array = f.readline().split()
                n_plies_S[j] = int(array[1])
                t_S[j] = float(array[2])
                theta_S[j] = float(array[3])
                mat_idx_S[j] = int(array[4]) - 1

            n_plies[i] = n_plies_S
            t[i] = t_S
            theta[i] = theta_S
            mat_idx[i] = mat_idx_S

        return n_plies, t, theta, mat_idx

    @staticmethod
    def __readWebLocFromFile(fname):
        # Get web locations from main input file
        f_main = os.path.join(os.path.split(fname)[0], os.path.split(fname)[1].replace("layup", "input"))

        # Error handling for different file extensions
        if not os.path.isfile(f_main):
            extensions = ["dat", "inp", "pci"]
            for ext in extensions:
                f_main = f_main[:-3] + ext
                if os.path.isfile(f_main):
                    break

        fid = open(f_main)

        var = fid.readline().split()[0]
        while var != "Web_num":
            text = fid.readline().split()
            if len(text) > 0:
                var = text[0]
            else:
                var = None

        web_loc = []
        line = fid.readline().split()
        while line:
            web_loc.append(float(line[1]))
            line = fid.readline().split()

        return web_loc

    def compositeMatrices(self, sector):
        """Computes the matrix components defining the constituitive equations
        of the complete laminate stack

        Returns
        -------
        A : ndarray, shape(3, 3)
            upper left portion of constitutive matrix
        B : ndarray, shape(3, 3)
            off-diagonal portion of constitutive matrix
        D : ndarray, shape(3, 3)
            lower right portion of constitutive matrix
        totalHeight : float (m)
            total height of the laminate stack

        Notes
        -----
        | The constitutive equations are arranged in the format
        | [N; M] = [A B; B D] * [epsilon; k]
        | where N = [N_x, N_y, N_xy] are the normal force resultants for the laminate
        | M = [M_x, M_y, M_xy] are the moment resultants
        | epsilon = [epsilon_x, epsilon_y, gamma_xy] are the midplane strains
        | k = [k_x, k_y, k_xy] are the midplane curvates

        See [1]_ for further details, and this :ref:`equation <ABBD>` in the user guide.

        References
        ----------
        .. [1] J. C. Halpin. Primer on Composite Materials Analysis. Technomic, 2nd edition, 1992.

        """

        t = self.t[sector]
        n_plies = self.n_plies[sector]
        mat_idx = self.mat_idx[sector]
        theta = self.theta[sector]

        mat_idx = mat_idx.astype(int)  # convert to integers if actually stored as floats

        n = len(theta)

        # heights (z - absolute, h - relative to mid-plane)
        z = np.zeros(n + 1)
        for i in range(n):
            z[i + 1] = z[i] + t[i] * n_plies[i]

        z_mid = (z[-1] - z[0]) / 2.0
        h = z - z_mid

        # ABD matrices
        A = np.zeros((3, 3))
        B = np.zeros((3, 3))
        D = np.zeros((3, 3))

        for i in range(n):
            Qbar = self.__Qbar(self.materials[mat_idx[i]], theta[i])
            A += Qbar * (h[i + 1] - h[i])
            B += 0.5 * Qbar * (h[i + 1] ** 2 - h[i] ** 2)
            D += 1.0 / 3.0 * Qbar * (h[i + 1] ** 3 - h[i] ** 3)

        totalHeight = z[-1] - z[0]

        return A, B, D, totalHeight

    def effectiveEAxial(self, sector):
        """Estimates the effective axial modulus of elasticity for the laminate

        Returns
        -------
        E : float (N/m^2)
            effective axial modulus of elasticity

        Notes
        -----
        see user guide for a :ref:`derivation <ABBD>`

        """

        A, B, D, totalHeight = self.compositeMatrices(sector)

        # S = [A B; B D]

        S = np.vstack((np.hstack((A, B)), np.hstack((B, D))))

        # E_eff_x = N_x/h/eps_xx and eps_xx = S^{-1}(0,0)*N_x (approximately)
        detS = np.linalg.det(S)
        Eaxial = detS / np.linalg.det(S[1:, 1:]) / totalHeight

        return Eaxial

    def __Qbar(self, material, theta):
        """Computes the lamina stiffness matrix

        Returns
        -------
        Qbar : numpy matrix
            the lamina stifness matrix

        Notes
        -----
        Transforms a specially orthotropic lamina from principal axis to
        an arbitrary axis defined by the ply orientation.
        [sigma_x; sigma_y; tau_xy]^T = Qbar * [epsilon_x; epsilon_y, gamma_xy]^T
        See [1]_ for further details.

        References
        ----------
        .. [1] J. C. Halpin. Primer on Composite Materials Analysis. Technomic, 2nd edition, 1992.


        """

        E11 = material.E1
        E22 = material.E2
        nu12 = material.nu12
        nu21 = nu12 * E22 / E11
        G12 = material.G12
        denom = 1 - nu12 * nu21

        c = math.cos(theta * math.pi / 180.0)
        s = math.sin(theta * math.pi / 180.0)
        c2 = c * c
        s2 = s * s
        cs = c * s

        Q = np.array([[E11 / denom, nu12 * E22 / denom, 0], [nu12 * E22 / denom, E22 / denom, 0], [0, 0, G12]])
        T12 = np.array([[c2, s2, cs], [s2, c2, -cs], [-cs, cs, 0.5 * (c2 - s2)]])
        Tinv = np.array([[c2, s2, -2 * cs], [s2, c2, 2 * cs], [cs, -cs, c2 - s2]])

        return Tinv @ Q @ T12

    def _preCompFormat(self):

        n = len(self.theta)
        n_lamina = np.zeros(n)

        if n == 0:
            return self.loc, n_lamina, self.n_plies, self.t, self.theta, self.mat_idx

        for i in range(n):
            n_lamina[i] = len(self.theta[i])

        mat = np.concatenate(self.mat_idx)
        for i in range(len(mat)):
            mat[i] += 1  # 1-based indexing in Fortran

        return self.loc, n_lamina, np.concatenate(self.n_plies), np.concatenate(self.t), np.concatenate(self.theta), mat


class Orthotropic2DMaterial:
    """Represents a homogeneous orthotropic material in a
    plane stress state.

    """

    def __init__(self, E1, E2, G12, nu12, rho, name=""):
        """a struct-like object.  all inputs are also fields.
        The object also has an identification
        number *.mat_idx so unique materials can be identified.

        Parameters
        ----------
        E1 : float (N/m^2)
            Young's modulus in first principal direction
        E2 : float (N/m^2)
            Young's modulus in second principal direction
        G12 : float (N/m^2)
            shear modulus
        nu12 : float
            Poisson's ratio  (nu12*E22 = nu21*E11)
        rho : float (kg/m^3)
            density
        name : str
            an optional identifier

        """
        self.E1 = E1
        self.E2 = E2
        self.G12 = G12
        self.nu12 = nu12
        self.rho = rho
        self.name = name

    @classmethod
    def listFromPreCompFile(cls, fname):
        """initialize the object by extracting materials from a PreComp materials file

        Parameters
        ----------
        fname : str
            name of input file

        Returns
        -------
        materials : List(:class:`Orthotropic2DMaterial`)
            a list of all materials gathered from the file.

        """

        f = open(fname)

        skipLines(f, 3)

        materials = []
        for line in f:
            array = line.split()
            mat = cls(float(array[1]), float(array[2]), float(array[3]), float(array[4]), float(array[5]), array[6])

            materials.append(mat)
        f.close()

        return materials


class Profile:
    """Defines the shape of an airfoil"""

    def __init__(self, xu, yu, xl, yl):
        """Constructor

        Parameters
        ----------
        xu : ndarray
            x coordinates for upper surface of airfoil
        yu : ndarray
            y coordinates for upper surface of airfoil
        xl : ndarray
            x coordinates for lower surface of airfoil
        yl : ndarray
            y coordinates for lower surface of airfoil

        Notes
        -----
        uses :ref:`airfoil-aligned coordinate system <blade_airfoil_coord>`.
        Nodes should be ordered from the leading edge toward the trailing edge.
        The leading edge can be located at any position and the
        chord may be any size, however the airfoil should be untwisted.
        Normalization to unit chord will happen internally.

        """

        # parse airfoil data
        xu = np.array(xu)
        yu = np.array(yu)
        xl = np.array(xl)
        yl = np.array(yl)

        # ensure leading edge at zero
        xu -= xu[0]
        xl -= xl[0]
        yu -= yu[0]
        yl -= yl[0]

        # ensure unit chord
        c = xu[-1] - xu[0]
        xu /= c
        xl /= c
        yu /= c
        yl /= c

        # interpolate onto common grid
        arc = np.linspace(0, math.pi, 100)
        self.x = 0.5 * (1 - np.cos(arc))  # cosine spacing
        self.yu = np.interp(self.x, xu, yu)
        self.yl = np.interp(self.x, xl, yl)

    @classmethod
    def initWithTEtoTEdata(cls, x, y):
        """Factory constructor for data points ordered from trailing edge to trailing edge.

        Parameters
        ----------
        x, y : ndarray, ndarray
            airfoil coordinates starting at trailing edge and
            ending at trailing edge, traversing airfoil in either direction.

        Returns
        -------
        profile : Profile
            initialized Profile object

        Notes
        -----
        It is not necessary to start and end at the same point
        for an airfoil with trailing edge thickness.
        Although, one point should be right at the nose.
        see also notes for :meth:`__init__`

        """

        # parse airfoil data
        x = np.array(x)
        y = np.array(y)

        # separate into 2 halves
        i = np.argmin(x)

        xu = x[i::-1]
        yu = y[i::-1]
        xl = x[i:]
        yl = y[i:]

        # check if coordinates were input in other direction
        if np.mean(y[0:i]) < np.mean(y[i:]):
            temp = yu
            yu = yl
            yl = temp

            temp = xu
            xu = xl
            xl = temp

        return cls(xu, yu, xl, yl)

    @classmethod
    def initWithLEtoLEdata(cls, x, y):
        """Factory constructor for data points ordered from leading edge to leading edge.

        Parameters
        ----------
        x, y : ndarray, ndarray
            airfoil coordinates starting at leading edge and
            ending at leading edge, traversing airfoil in either direction.

        Returns
        -------
        profile : Profile
            initialized Profile object

        Notes
        -----
        x,y data must start and end at the same point.
        see also notes for :meth:`__init__`

        """

        # parse airfoil data
        x = np.array(x)
        y = np.array(y)

        # separate into 2 halves
        for i in range(len(x)):
            if x[i + 1] <= x[i]:
                iuLast = i
                ilLast = i
                if x[i + 1] == x[i]:  # blunt t.e.
                    ilLast = i + 1  # stop at i+1
                break

        xu = x[: iuLast + 1]
        yu = y[: iuLast + 1]
        xl = x[-1 : ilLast - 1 : -1]
        yl = y[-1 : ilLast - 1 : -1]

        # check if coordinates were input in other direction
        if y[1] < y[0]:
            temp = yu
            yu = yl
            yl = temp

            temp = xu
            xu = xl
            xl = temp

        return cls(xu, yu, xl, yl)

    @staticmethod
    def initFromPreCompFile(precompProfileFile):
        """Construct profile from PreComp formatted file

        Parameters
        ----------
        precompProfileFile : str
            path/name of file

        Returns
        -------
        profile : Profile
            initialized Profile object

        """

        return Profile.initFromFile(precompProfileFile, 4, True)

    @staticmethod
    def initFromFile(filename, numHeaderlines, LEtoLE):
        """Construct profile from a generic form text file (see Notes)

        Parameters
        ----------
        filename : str
            name/path of input file
        numHeaderlines : int
            number of header rows in input file
        LEtoLE : boolean
            True if data is ordered from leading-edge to leading-edge
            False if from trailing-edge to trailing-edge

        Returns
        -------
        profile : Profile
            initialized Profile object

        Notes
        -----
        file should be of the form:

        header row
        header row
        x1 y1
        x2 y2
        x3 y3
        .  .
        .  .
        .  .

        where any number of header rows can be used.

        """

        # open file
        f = open(filename, "r")

        # skip through header
        for i in range(numHeaderlines):
            f.readline()

        # loop through
        x = []
        y = []

        for line in f:
            if not line.strip():
                break  # break if empty line
            data = line.split()
            x.append(float(data[0]))
            y.append(float(data[1]))

        f.close()

        # close nose if LE to LE
        if LEtoLE:
            x.append(x[0])
            y.append(y[0])
            return Profile.initWithLEtoLEdata(x, y)

        else:
            return Profile.initWithTEtoTEdata(x, y)

    def _preCompFormat(self):
        """
        docstring
        """

        # check if they share a common trailing edge point
        te_same = self.yu[-1] == self.yl[-1]

        # count number of points
        nu = len(self.x)
        if te_same:
            nu -= 1
        nl = len(self.x) - 1  # they do share common leading-edge
        n = nu + nl

        # initialize
        x = np.zeros(n)
        y = np.zeros(n)

        # leading edge round to leading edge
        x[0:nu] = self.x[0:nu]
        y[0:nu] = self.yu[0:nu]
        x[nu:] = self.x[:0:-1]
        y[nu:] = self.yl[:0:-1]

        return x, y

    def locationOfMaxThickness(self):
        """Find location of max airfoil thickness

        Returns
        -------
        x : float
            x location of maximum thickness
        yu : float
            upper surface y location of maximum thickness
        yl : float
            lower surface y location of maximum thickness

        Notes
        -----
        uses :ref:`airfoil-aligned coordinate system <blade_airfoil_coord>`

        """

        idx = np.argmax(self.yu - self.yl)
        return (self.x[idx], self.yu[idx], self.yl[idx])

    def blend(self, other, weight):
        """Blend this profile with another one with the specified weighting.

        Parameters
        ----------
        other : Profile
            another Profile to blend with
        weight : float
            blending parameter between 0 and 1.  0 returns self, whereas 1 returns other.

        Returns
        -------
        profile : Profile
            a blended profile

        """

        # blend coordinates
        yu = self.yu + weight * (other.yu - self.yu)
        yl = self.yl + weight * (other.yl - self.yl)

        return Profile(self.x, yu, self.x, yl)

    @property
    def tc(self):
        """thickness to chord ratio of the Profile"""
        return max(self.yu - self.yl)

    def set_tc(self, new_tc):

        factor = new_tc / self.tc

        self.yu *= factor
        self.yl *= factor


class PreCompWriter:
    def __init__(self, dir_out, materials, upper, lower, webs, profile, chord, twist, p_le):
        self.dir_out = dir_out

        self.materials = materials
        self.upper = upper
        self.lower = lower
        self.webs = webs
        self.profile = profile

        self.chord = chord
        self.twist = twist
        self.p_le = p_le

        if not os.path.exists(dir_out):
            os.makedirs(dir_out)

    def execute(self):

        flist_layup = self.writePreCompLayup()
        flist_profile = self.writePreCompProfile()
        self.writePreCompMaterials()
        self.writePreCompInput(flist_layup, flist_profile)

    def writePreCompMaterials(self):

        text = []
        text.append("\n")
        text.append("Mat_Id     E1           E2          G12       Nu12     Density      Mat_Name\n")
        text.append(" (-)      (Pa)         (Pa)        (Pa)       (-)      (Kg/m^3)       (-)\n")

        for i, mat in enumerate(self.materials):
            text.append("%d %e %e %e %f %e %s\n" % (i + 1, mat.E1, mat.E2, mat.G12, mat.nu12, mat.rho, mat.name))

        fout = os.path.join(self.dir_out, "materials.inp")
        f = open(fout, "w")
        for outLine in text:
            f.write(outLine)
        f.close()

    def writePreCompLayup(self):
        f_out = []

        def write_layup_sectors(cs, web):
            text = []
            for i, (n_plies, t, theta, mat_idx) in enumerate(zip(cs.n_plies, cs.t, cs.theta, cs.mat_idx)):
                if web:
                    text.extend(["\n", "web_num    no of laminae (N_weblams)    Name of stack:\n"])
                else:
                    text.extend(
                        [
                            "...........................................................................\n",
                            "Sect_num    no of laminae (N_laminas)          STACK:\n",
                        ]
                    )
                n_lamina = len(n_plies)
                text.append("%d %d\n" % (i + 1, n_lamina))
                text.extend(
                    [
                        "\n",
                        "lamina    num of  thickness   fibers_direction  composite_material ID\n",
                        "number    plies   of ply (m)       (deg)               (-)\n",
                    ]
                )
                if web:
                    text.append("wlam_num N_Plies   w_tply       Tht_Wlam            Wmat_Id\n")
                else:
                    text.append("lam_num  N_plies    Tply         Tht_lam            Mat_id\n")

                for j, (plies_j, t_j, theta_j, mat_idx_j) in enumerate(zip(n_plies, t, theta, mat_idx + 1)):
                    text.append("%d %d %e %.1f %d\n" % (j + 1, plies_j, t_j, theta_j, mat_idx_j))
            return text

        for idx, (lower_i, upper_i, webs_i) in enumerate(zip(self.lower, self.upper, self.webs)):

            text = []
            text.append("Composite laminae lay-up inside the blade section\n")
            text.append("\n")
            text.append("*************************** TOP SURFACE ****************************\n")
            # number of sectors
            n_sector = len(upper_i.loc) - 1
            text.append("%d                N_scts(1):  no of sectors on top surface\n" % n_sector)
            text.extend(["\n", "normalized chord location of  nodes defining airfoil sectors boundaries (xsec_node)\n"])
            locU = upper_i.loc
            text.append(" ".join(["%f" % i for i in locU]) + "\n")
            text.extend(write_layup_sectors(upper_i, False))

            text.extend(["\n", "\n", "*************************** BOTTOM SURFACE ****************************\n"])
            n_sector = len(lower_i.loc) - 1
            text.append("%d                N_scts(1):  no of sectors on top surface\n" % n_sector)
            text.extend(["\n", "normalized chord location of  nodes defining airfoil sectors boundaries (xsec_node)\n"])
            locU = lower_i.loc
            text.append(" ".join(["%f" % i for i in locU]) + "\n")
            text.extend(write_layup_sectors(lower_i, False))

            text.extend(
                [
                    "\n",
                    "\n",
                    "**********************************************************************\n",
                    "Laminae schedule for webs (input required only if webs exist at this section):\n",
                ]
            )
            ########## Webs ##########
            text.extend(write_layup_sectors(webs_i, True))

            fname = os.path.join(self.dir_out, "layup_%00d.inp" % idx)
            f = open(fname, "w")
            for outLine in text:
                f.write(outLine)
            f.close()
            f_out.append(fname)

        return f_out

    def writePreCompProfile(self):
        f_out = []
        for idx, profile_i in enumerate(self.profile):
            # idx = 0
            # profile_i = profile[idx]
            text = []

            text.append(
                "%d                      N_af_nodes :no of airfoil nodes, counted clockwise starting\n"
                % len(profile_i.x)
            )
            text.append("                      with leading edge (see users' manual, fig xx)\n")
            text.append("\n")
            text.append(" Xnode      Ynode   !! chord-normalized coordinated of the airfoil nodes\n")

            x_all = np.concatenate((profile_i.x, np.flip(profile_i.x, 0)))
            y_all = np.concatenate((profile_i.yu, np.flip(profile_i.yl, 0)))

            if max(y_all) > 1.0:
                print(idx)

            # import matplotlib.pyplot as plt
            # plt.figure()
            # plt.plot(x_all, y_all)
            # plt.savefig('test.png')

            for x, y in zip(x_all, y_all):
                text.append("%f %f\n" % (x, y))

            fname = os.path.join(self.dir_out, "shape_%d.inp" % idx)
            f = open(fname, "w")
            for outLine in text:
                f.write(outLine)
            f.close()
            f_out.append(fname)

        return f_out

    def writePreCompInput(self, flist_layup, flist_profile):

        for idx in range(0, len(flist_layup)):

            chord = self.chord[idx]
            twist = self.twist[idx]
            p_le = self.p_le[idx]
            webs_i = self.webs[idx]
            layup_i = os.path.split(os.path.abspath(flist_layup[idx]))[1]
            profile_i = os.path.split(os.path.abspath(flist_profile[idx]))[1]

            text = []
            text.append("*****************  main input file for PreComp *****************************\n")
            text.append("Sample Composite Blade Section Properties\n")
            text.append("\n")
            text.append("General information -----------------------------------------------\n")
            text.append("1                Bl_length   : blade length (m)\n")
            text.append("2                N_sections  : no of blade sections (-)\n")
            text.append(
                "%d                N_materials : no of materials listed in the materials table (material.inp)\n"
                % len(self.materials)
            )
            text.append("3                Out_format  : output file   (1: general format, 2: BModes-format, 3: both)\n")
            text.append("f                TabDelim     (true: tab-delimited table; false: space-delimited table)\n")
            text.append("\n")
            text.append("Blade-sections-specific data --------------------------------------\n")
            text.append("Sec span     l.e.     chord   aerodynamic   af_shape    int str layup\n")
            text.append("location   position   length    twist         file          file\n")
            text.append("Span_loc    Le_loc    Chord    Tw_aero   Af_shape_file  Int_str_file\n")
            text.append("  (-)        (-)       (m)    (degrees)       (-)           (-)\n")
            text.append("\n")
            text.append("%.2f %f %e %f %s %s\n" % (0.0, p_le, chord, twist, profile_i, layup_i))
            text.append("%.2f %f %e %f %s %s\n" % (1.0, p_le, chord, twist, profile_i, layup_i))
            text.append("\n")
            text.append("Webs (spars) data  --------------------------------------------------\n")
            text.append("\n")
            text.append(
                "%d                Nweb        : number of webs (-)  ! enter 0 if the blade has no webs\n"
                % len(webs_i.loc)
            )
            text.append(
                "1                Ib_sp_stn   : blade station number where inner-most end of webs is located (-)\n"
            )
            text.append(
                "2                Ob_sp_stn   : blade station number where outer-most end of webs is located (-)\n"
            )
            text.append("\n")
            text.append("Web_num   Inb_end_ch_loc   Oub_end_ch_loc (fraction of chord length)\n")
            for i, loc in enumerate(webs_i.loc):
                text.append("%d %f %f\n" % (i + 1, loc, loc))

            fname = os.path.join(self.dir_out, "input_%d.inp" % idx)
            f = open(fname, "w")
            for outLine in text:
                f.write(outLine)
            f.close()


if __name__ == "__main__":

    import os

    # geometry
    r_str = [
        1.5,
        1.80135,
        1.89975,
        1.99815,
        2.1027,
        2.2011,
        2.2995,
        2.87145,
        3.0006,
        3.099,
        5.60205,
        6.9981,
        8.33265,
        10.49745,
        11.75205,
        13.49865,
        15.84795,
        18.4986,
        19.95,
        21.99795,
        24.05205,
        26.1,
        28.14795,
        32.25,
        33.49845,
        36.35205,
        38.4984,
        40.44795,
        42.50205,
        43.49835,
        44.55,
        46.49955,
        48.65205,
        52.74795,
        56.16735,
        58.89795,
        61.62855,
        63.0,
    ]
    chord_str = [
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.387,
        3.39,
        3.741,
        4.035,
        4.25,
        4.478,
        4.557,
        4.616,
        4.652,
        4.543,
        4.458,
        4.356,
        4.249,
        4.131,
        4.007,
        3.748,
        3.672,
        3.502,
        3.373,
        3.256,
        3.133,
        3.073,
        3.01,
        2.893,
        2.764,
        2.518,
        2.313,
        2.086,
        1.419,
        1.085,
    ]
    theta_str = [
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        12.53,
        11.48,
        10.63,
        10.16,
        9.59,
        9.01,
        8.4,
        7.79,
        6.54,
        6.18,
        5.36,
        4.75,
        4.19,
        3.66,
        3.4,
        3.13,
        2.74,
        2.32,
        1.53,
        0.86,
        0.37,
        0.11,
        0.0,
    ]
    le_str = [
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.498,
        0.497,
        0.465,
        0.447,
        0.43,
        0.411,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
    ]
    web1 = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            0.4114,
            0.4102,
            0.4094,
            0.3876,
            0.3755,
            0.3639,
            0.345,
            0.3342,
            0.3313,
            0.3274,
            0.323,
            0.3206,
            0.3172,
            0.3138,
            0.3104,
            0.307,
            0.3003,
            0.2982,
            0.2935,
            0.2899,
            0.2867,
            0.2833,
            0.2817,
            0.2799,
            0.2767,
            0.2731,
            0.2664,
            0.2607,
            0.2562,
            0.1886,
            -1.0,
        ]
    )
    web2 = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            0.5886,
            0.5868,
            0.5854,
            0.5508,
            0.5315,
            0.5131,
            0.4831,
            0.4658,
            0.4687,
            0.4726,
            0.477,
            0.4794,
            0.4828,
            0.4862,
            0.4896,
            0.493,
            0.4997,
            0.5018,
            0.5065,
            0.5101,
            0.5133,
            0.5167,
            0.5183,
            0.5201,
            0.5233,
            0.5269,
            0.5336,
            0.5393,
            0.5438,
            0.6114,
            -1.0,
        ]
    )
    web3 = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
        ]
    )
    precurve_str = np.zeros_like(r_str)
    presweep_str = np.zeros_like(r_str)

    # -------- materials and composite layup  -----------------
    basepath = os.path.join("5MW_files", "5MW_PreCompFiles")

    materials = Orthotropic2DMaterial.listFromPreCompFile(os.path.join(basepath, "materials.inp"))

    ncomp = len(r_str)
    upper = [0] * ncomp
    lower = [0] * ncomp
    webs = [0] * ncomp
    profile = [0] * ncomp

    # # web 1
    # ib_idx = 7
    # ob_idx = 36
    # ib_webc = 0.4114
    # ob_webc = 0.1886

    # web1 = web_loc(r_str, chord_str, le_str, ib_idx, ob_idx, ib_webc, ob_webc)

    # # web 2
    # ib_idx = 7
    # ob_idx = 36
    # ib_webc = 0.5886
    # ob_webc = 0.6114

    # web2 = web_loc(r_str, chord_str, le_str, ib_idx, ob_idx, ib_webc, ob_webc)

    for i in range(ncomp):

        webLoc = []
        if web1[i] != -1:
            webLoc.append(web1[i])
        if web2[i] != -1:
            webLoc.append(web2[i])
        if web3[i] != -1:
            webLoc.append(web3[i])

        upper[i], lower[i], webs[i] = CompositeSection.initFromPreCompLayupFile(
            os.path.join(basepath, "layup_" + str(i + 1) + ".inp"), webLoc, materials
        )
        profile[i] = Profile.initFromPreCompFile(os.path.join(basepath, "shape_" + str(i + 1) + ".inp"))
    # --------------------------------------

    precomp = PreComp(
        r_str, chord_str, theta_str, le_str, precurve_str, presweep_str, profile, materials, upper, lower, webs
    )

    # evalute section properties
    EA, EIxx, EIyy, EIxy, GJ, rhoA, rhoJ, x_ec_str, y_ec_str = precomp.sectionProperties()

    import matplotlib.pyplot as plt

    r_str = np.array(r_str)
    rstar = (r_str - r_str[0]) / (r_str[-1] - r_str[0])

    plt.figure(1)
    plt.semilogy(rstar, EIxx)
    plt.xlabel("blade fraction")
    plt.ylabel("Edgewise Stiffness ($N m^2$)")

    plt.figure(2)
    plt.semilogy(rstar, EIyy)
    plt.xlabel("blade fraction")
    plt.ylabel("Flapwise Stiffness ($N m^2$)")

    plt.figure(3)
    plt.semilogy(rstar, EA)
    plt.figure(4)
    plt.semilogy(rstar, EIxy)
    plt.figure(5)
    plt.semilogy(rstar, GJ)
    plt.figure(6)
    plt.semilogy(rstar, rhoA)
    plt.figure(7)
    plt.semilogy(rstar, rhoJ)
    plt.figure(8)
    plt.plot(rstar, x_ec_str)
    plt.figure(9)
    plt.plot(rstar, y_ec_str)
    plt.figure(10)
    plt.plot(rstar, precomp.x_ec_nose)
    plt.figure(11)
    plt.plot(rstar, precomp.y_ec_nose)

    plt.show()



import numpy as np
import openmdao.api as om

import wisdem.commonse.utilities as util
import wisdem.pyframe3dd.pyframe3dd as pyframe3dd
import wisdem.commonse.utilization_dnvgl as util_dnvgl
import wisdem.commonse.utilization_constraints as util_con
from wisdem.commonse import NFREQ, gravity
from wisdem.commonse.cylinder_member import NULL, MEMMAX, MemberLoads, get_nfull
from wisdem.floatingse.floating_system import RIGID, NELEM_MAX, NNODES_MAX


class PlatformLoads(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_dlc = opt["WISDEM"]["n_dlc"]
        n_member = opt["floating"]["members"]["n_members"]

        for k in range(n_member):
            self.add_input(f"member{k}:Px", np.zeros((MEMMAX, n_dlc)), units="N/m")
            self.add_input(f"member{k}:Py", np.zeros((MEMMAX, n_dlc)), units="N/m")
            self.add_input(f"member{k}:Pz", np.zeros((MEMMAX, n_dlc)), units="N/m")
            self.add_input(f"member{k}:qdyn", np.zeros((MEMMAX, n_dlc)), units="Pa")

        self.add_output("platform_elem_Px1", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_output("platform_elem_Px2", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_output("platform_elem_Py1", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_output("platform_elem_Py2", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_output("platform_elem_Pz1", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_output("platform_elem_Pz2", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_output("platform_elem_qdyn", NULL * np.ones((NELEM_MAX, n_dlc)), units="Pa")

    def compute(self, inputs, outputs):
        # Load in number of members
        opt = self.options["options"]
        n_dlc = opt["WISDEM"]["n_dlc"]
        n_member = opt["floating"]["members"]["n_members"]

        # Initialize running lists across all members
        elem_Px1 = np.array([])
        elem_Px2 = np.array([])
        elem_Py1 = np.array([])
        elem_Py2 = np.array([])
        elem_Pz1 = np.array([])
        elem_Pz2 = np.array([])
        elem_qdyn = np.array([])

        # Append all member data
        for k in range(n_member):
            n = np.where(inputs[f"member{k}:qdyn"][:, 0] == NULL)[0][0]
            mem_qdyn, _ = util.nodal2sectional(inputs[f"member{k}:qdyn"][:n, :])

            # The loads should come in with length n+1
            n -= 1
            if k == 0:
                elem_Px1 = inputs[f"member{k}:Px"][:n, :]
                elem_Px2 = inputs[f"member{k}:Px"][1 : (n + 1), :]
                elem_Py1 = inputs[f"member{k}:Py"][:n, :]
                elem_Py2 = inputs[f"member{k}:Py"][1 : (n + 1), :]
                elem_Pz1 = inputs[f"member{k}:Pz"][:n, :]
                elem_Pz2 = inputs[f"member{k}:Pz"][1 : (n + 1), :]
                elem_qdyn = mem_qdyn
            else:
                elem_Px1 = np.vstack((elem_Px1, inputs[f"member{k}:Px"][:n, :]))
                elem_Px2 = np.vstack((elem_Px2, inputs[f"member{k}:Px"][1 : (n + 1), :]))
                elem_Py1 = np.vstack((elem_Py1, inputs[f"member{k}:Py"][:n, :]))
                elem_Py2 = np.vstack((elem_Py2, inputs[f"member{k}:Py"][1 : (n + 1), :]))
                elem_Pz1 = np.vstack((elem_Pz1, inputs[f"member{k}:Pz"][:n, :]))
                elem_Pz2 = np.vstack((elem_Pz2, inputs[f"member{k}:Pz"][1 : (n + 1), :]))
                elem_qdyn = np.vstack((elem_qdyn, mem_qdyn))

        # Store outputs
        nelem = elem_qdyn.size
        outputs["platform_elem_Px1"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_elem_Px2"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_elem_Py1"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_elem_Py2"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_elem_Pz1"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_elem_Pz2"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_elem_qdyn"] = NULL * np.ones((NELEM_MAX, n_dlc))

        outputs["platform_elem_Px1"][:nelem, :] = elem_Px1
        outputs["platform_elem_Px2"][:nelem, :] = elem_Px2
        outputs["platform_elem_Py1"][:nelem, :] = elem_Py1
        outputs["platform_elem_Py2"][:nelem, :] = elem_Py2
        outputs["platform_elem_Pz1"][:nelem, :] = elem_Pz1
        outputs["platform_elem_Pz2"][:nelem, :] = elem_Pz2
        outputs["platform_elem_qdyn"][:nelem, :] = elem_qdyn


class FrameAnalysis(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_attach = opt["mooring"]["n_attach"]
        n_dlc = opt["WISDEM"]["n_dlc"]

        self.add_input("platform_mass", 0.0, units="kg")
        self.add_input("platform_hull_center_of_mass", np.zeros(3), units="m")
        self.add_input("platform_added_mass", np.zeros(6), units="kg")

        self.add_input("platform_nodes", NULL * np.ones((NNODES_MAX, 3)), units="m")
        self.add_input("platform_Fnode", NULL * np.ones((NNODES_MAX, 3)), units="N")
        self.add_input("platform_Rnode", NULL * np.ones(NNODES_MAX), units="m")
        self.add_input("platform_elem_n1", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_input("platform_elem_n2", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_input("platform_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_L", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_J0", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_rho", NULL * np.ones(NELEM_MAX), units="kg/m**3")
        self.add_input("platform_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")

        self.add_input("platform_elem_Px1", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_input("platform_elem_Px2", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_input("platform_elem_Py1", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_input("platform_elem_Py2", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_input("platform_elem_Pz1", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")
        self.add_input("platform_elem_Pz2", NULL * np.ones((NELEM_MAX, n_dlc)), units="N/m")

        self.add_input("transition_node", np.zeros(3), units="m")
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_I", np.zeros(6), units="kg*m**2")
        self.add_input("turbine_F", np.zeros((3, n_dlc)), units="N")
        self.add_input("turbine_M", np.zeros((3, n_dlc)), units="N*m")
        self.add_input("mooring_neutral_load", np.zeros((n_attach, 3)), units="N")
        self.add_input("mooring_fairlead_joints", np.zeros((n_attach, 3)), units="m")
        self.add_input("mooring_stiffness", np.zeros((6, 6)), units="N/m")
        self.add_input("variable_ballast_mass", 0.0, units="kg")
        self.add_input("variable_center_of_mass", val=np.zeros(3), units="m")
        self.add_input("variable_I", np.zeros(6), units="kg*m**2")

        self.add_output("platform_base_F", np.zeros((3, n_dlc)), units="N")
        self.add_output("platform_base_M", np.zeros((3, n_dlc)), units="N*m")
        self.add_output("platform_Fz", NULL * np.ones((NELEM_MAX, n_dlc)), units="N")
        self.add_output("platform_Vx", NULL * np.ones((NELEM_MAX, n_dlc)), units="N")
        self.add_output("platform_Vy", NULL * np.ones((NELEM_MAX, n_dlc)), units="N")
        self.add_output("platform_Mxx", NULL * np.ones((NELEM_MAX, n_dlc)), units="N*m")
        self.add_output("platform_Myy", NULL * np.ones((NELEM_MAX, n_dlc)), units="N*m")
        self.add_output("platform_Mzz", NULL * np.ones((NELEM_MAX, n_dlc)), units="N*m")

    def compute(self, inputs, outputs):

        # Unpack variables
        opt = self.options["options"]
        n_attach = opt["mooring"]["n_attach"]
        n_dlc = opt["WISDEM"]["n_dlc"]

        m_trans = float(inputs["transition_piece_mass"])
        I_trans = inputs["transition_piece_I"]
        m_variable = float(inputs["variable_ballast_mass"])
        cg_variable = inputs["variable_center_of_mass"]
        I_variable = inputs["variable_I"]

        fairlead_joints = inputs["mooring_fairlead_joints"]
        mooringF = inputs["mooring_neutral_load"]
        mooringK = np.abs(np.diag(inputs["mooring_stiffness"]))

        # Create frame3dd instance: nodes, elements, reactions, and options
        nodes = inputs["platform_nodes"]
        nnode = np.where(nodes[:, 0] == NULL)[0][0]
        nodes = nodes[:nnode, :]
        rnode = np.zeros(nnode)  # inputs["platform_Rnode"][:nnode]
        itrans = util.closest_node(nodes, inputs["transition_node"])
        ivariable = util.closest_node(nodes, cg_variable)

        N1 = np.int_(inputs["platform_elem_n1"])
        nelem = np.where(N1 == NULL)[0][0]
        N1 = N1[:nelem]
        N2 = np.int_(inputs["platform_elem_n2"][:nelem])
        A = inputs["platform_elem_A"][:nelem]
        Asx = inputs["platform_elem_Asx"][:nelem]
        Asy = inputs["platform_elem_Asy"][:nelem]
        Ixx = inputs["platform_elem_Ixx"][:nelem]
        Iyy = inputs["platform_elem_Iyy"][:nelem]
        J0 = inputs["platform_elem_J0"][:nelem]
        rho = inputs["platform_elem_rho"][:nelem]
        E = inputs["platform_elem_E"][:nelem]
        G = inputs["platform_elem_G"][:nelem]
        roll = np.zeros(nelem)
        L = inputs["platform_elem_L"][:nelem]  # np.sqrt(np.sum((nodes[N2, :] - nodes[N1, :]) ** 2, axis=1))

        inodes = np.arange(nnode) + 1
        node_obj = pyframe3dd.NodeData(inodes, nodes[:, 0], nodes[:, 1], nodes[:, 2], rnode)

        ielem = np.arange(nelem) + 1
        elem_obj = pyframe3dd.ElementData(ielem, N1 + 1, N2 + 1, A, Asx, Asy, J0, Ixx, Iyy, E, G, roll, rho)

        # Use Mooring stiffness (TODO Hydro_K too)
        ind = []
        for k in range(n_attach):
            ind.append(util.closest_node(nodes, fairlead_joints[k, :]))
        rid = np.unique(np.array([ind]))  # np.array([np.argmin(nodes[:, 2])])

        Rx = Ry = Rz = Rxx = Ryy = Rzz = RIGID * np.ones(rid.size)
        # Rx, Ry, Rz = [mooringK[0]], [mooringK[1]], [mooringK[2]]
        # Only this solution works and there isn't much different with fully rigid
        # Rx, Ry, Rz = [RIGID], [RIGID], [mooringK[2]]
        # Rxx, Ryy, Rzz = [RIGID], [RIGID], [RIGID]
        react_obj = pyframe3dd.ReactionData(rid + 1, Rx, Ry, Rz, Rxx, Ryy, Rzz, rigid=RIGID)

        frame3dd_opt = opt["WISDEM"]["FloatingSE"]["frame3dd"]
        opt_obj = pyframe3dd.Options(frame3dd_opt["shear"], frame3dd_opt["geom"], -1.0)

        myframe = pyframe3dd.Frame(node_obj, react_obj, elem_obj, opt_obj)

        # Added mass
        cg_add = np.zeros((3, 2))
        add_gravity = True
        mID = np.array([itrans, ivariable], dtype=np.int_)
        m_add = np.array([m_trans, m_variable])
        I_add = np.c_[I_trans, I_variable]
        myframe.changeExtraNodeMass(
            mID + 1,
            m_add,
            I_add[0, :],
            I_add[1, :],
            I_add[2, :],
            I_add[3, :],
            I_add[4, :],
            I_add[5, :],
            cg_add[0, :],
            cg_add[1, :],
            cg_add[2, :],
            add_gravity,
        )

        # Initialize loading with gravity, mooring line forces, and buoyancy (already in nodal forces)
        Mnode = np.zeros((nnode, 3))
        Fnode = inputs["platform_Fnode"][:nnode, :]
        for k in range(n_attach):
            ind = util.closest_node(nodes, fairlead_joints[k, :])
            Fnode[ind, :] += mooringF[k, :]

        gx = gy = 0.0
        gz = -gravity
        for k in range(n_dlc):
            load_obj = pyframe3dd.StaticLoadCase(gx, gy, gz)

            Fnode2 = Fnode.copy()
            Fnode2[itrans, :] += inputs["turbine_F"][:, k]
            Mnode[itrans, :] = inputs["turbine_M"][:, k]
            nF = np.where(np.abs(Fnode2).sum(axis=1) > 0.0)[0]
            load_obj.changePointLoads(
                nF + 1, Fnode2[nF, 0], Fnode2[nF, 1], Fnode2[nF, 2], Mnode[nF, 0], Mnode[nF, 1], Mnode[nF, 2]
            )

            # trapezoidally distributed loads
            xx1 = xy1 = xz1 = np.zeros(ielem.size)
            xx2 = xy2 = xz2 = 0.99 * L  # multiply slightly less than unity b.c. of precision
            wx1 = inputs["platform_elem_Px1"][:nelem, k]
            wx2 = inputs["platform_elem_Px2"][:nelem, k]
            wy1 = inputs["platform_elem_Py1"][:nelem, k]
            wy2 = inputs["platform_elem_Py2"][:nelem, k]
            wz1 = inputs["platform_elem_Pz1"][:nelem, k]
            wz2 = inputs["platform_elem_Pz2"][:nelem, k]
            load_obj.changeTrapezoidalLoads(ielem, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

            # Add the load case and run
            myframe.addLoadCase(load_obj)

        # myframe.write("system.3dd")
        # myframe.draw()
        displacements, forces, reactions, internalForces, mass, modal = myframe.run()

        # Determine reaction forces
        outputs["platform_base_F"] = -np.c_[
            reactions.Fx.sum(axis=1), reactions.Fy.sum(axis=1), reactions.Fz.sum(axis=1)
        ].T
        outputs["platform_base_M"] = -np.c_[
            reactions.Mxx.sum(axis=1), reactions.Myy.sum(axis=1), reactions.Mzz.sum(axis=1)
        ].T

        # Forces and moments along the structure
        outputs["platform_Fz"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_Vx"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_Vy"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_Mxx"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_Myy"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["platform_Mzz"] = NULL * np.ones((NELEM_MAX, n_dlc))
        for k in range(n_dlc):
            outputs["platform_Fz"][:nelem, k] = forces.Nx[k, 1::2]
            outputs["platform_Vx"][:nelem, k] = -forces.Vz[k, 1::2]
            outputs["platform_Vy"][:nelem, k] = forces.Vy[k, 1::2]
            outputs["platform_Mxx"][:nelem, k] = -forces.Mzz[k, 1::2]
            outputs["platform_Myy"][:nelem, k] = forces.Myy[k, 1::2]
            outputs["platform_Mzz"][:nelem, k] = forces.Txx[k, 1::2]


class TowerModal(om.ExplicitComponent):
    """
    Run Frame3DD on the floating tower for frequencies and mode shapes only

    Parameters
    ----------
    z_full : numpy array[npts], [m]
        location along cylinder. start at bottom and go to top
    d_full : numpy array[npts], [m]
        effective cylinder diameter for section
    t_full : numpy array[npts-1], [m]
        effective shell thickness for section
    E_full : numpy array[npts-1], [N/m**2]
        modulus of elasticity
    G_full : numpy array[npts-1], [N/m**2]
        shear modulus
    rho_full : numpy array[npts-1], [kg/m**3]
        material density

    Returns
    -------
    f1 : float, [Hz]
        First natural frequency
    f2 : float, [Hz]
        Second natural frequency
    structural_frequencies : numpy array[NFREQ], [Hz]
        First and second natural frequency
    fore_aft_freqs : numpy array[NFREQ2]
        Frequencies associated with mode shapes in the tower fore-aft direction
    side_side_freqs : numpy array[NFREQ2]
        Frequencies associated with mode shapes in the tower side-side direction
    torsion_freqs : numpy array[NFREQ2]
        Frequencies associated with mode shapes in the tower torsion direction
    fore_aft_modes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the tower fore-aft direction
        (without constant term)
    side_side_modes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the tower side-side direction
        (without constant term)
    torsion_modes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the tower torsion direction
        (without constant term)
    """

    def initialize(self):
        self.options.declare("n_full")
        self.options.declare("frame3dd_opt")
        self.options.declare("rank_and_file", default=False)

    def setup(self):
        n_full = self.options["n_full"]

        # cross-sectional data along cylinder.
        self.add_input("tower_xyz", np.zeros((n_full, 3)), units="m")
        self.add_input("tower_A", np.zeros(n_full - 1), units="m**2")
        self.add_input("tower_Asx", np.zeros(n_full - 1), units="m**2")
        self.add_input("tower_Asy", np.zeros(n_full - 1), units="m**2")
        self.add_input("tower_Ixx", np.zeros(n_full - 1), units="kg*m**2")
        self.add_input("tower_Iyy", np.zeros(n_full - 1), units="kg*m**2")
        self.add_input("tower_J0", np.zeros(n_full - 1), units="kg*m**2")
        self.add_input("tower_rho", np.zeros(n_full - 1), units="kg/m**3")
        self.add_input("tower_E", np.zeros(n_full - 1), units="Pa")
        self.add_input("tower_G", np.zeros(n_full - 1), units="Pa")
        self.add_output("tower_L", np.zeros(n_full - 1), units="m")

        self.add_input("rna_mass", val=0.0, units="kg")
        self.add_input("rna_I", np.zeros(6), units="kg*m**2")
        self.add_input("rna_cg", np.zeros(3), units="m")
        self.add_input("platform_mass", 0.0, units="kg")
        self.add_input("platform_added_mass", np.zeros(6), units="kg")
        self.add_input("platform_total_center_of_mass", np.zeros(3), units="m")
        self.add_input("platform_I_total", np.zeros(6), units="kg*m**2")
        self.add_input("mooring_stiffness", np.zeros((6, 6)), units="N/m")

        # Frequencies
        NFREQ2 = int(NFREQ / 2)
        self.add_output("f1", val=0.0, units="Hz")
        self.add_output("f2", val=0.0, units="Hz")
        self.add_output("structural_frequencies", np.zeros(NFREQ), units="Hz")
        self.add_output("fore_aft_modes", np.zeros((NFREQ2, 5)))
        self.add_output("side_side_modes", np.zeros((NFREQ2, 5)))
        self.add_output("torsion_modes", np.zeros((NFREQ2, 5)))
        self.add_output("fore_aft_freqs", np.zeros(NFREQ2), units="Hz")
        self.add_output("side_side_freqs", np.zeros(NFREQ2), units="Hz")
        self.add_output("torsion_freqs", np.zeros(NFREQ2), units="Hz")

    def compute(self, inputs, outputs):
        frame3dd_opt = self.options["frame3dd_opt"]

        # ------- node data ----------------
        xyz = inputs["tower_xyz"]
        n = xyz.shape[0]
        node = np.arange(1, n + 1)
        r = np.zeros(n)
        nodes = pyframe3dd.NodeData(node, xyz[:, 0], xyz[:, 1], xyz[:, 2], r)
        # -----------------------------------

        # ------ reaction data ------------
        # free-free (no reactions)
        rnode = np.array([], dtype=np.int_)
        kx = ky = kz = ktx = kty = ktz = rnode
        reactions = pyframe3dd.ReactionData(rnode, kx, ky, kz, ktx, kty, ktz, rigid=RIGID)
        # rnode = np.array([1], dtype=np.int_)
        # moorK = np.abs(np.diag(inputs["mooring_stiffness"]))
        # reactions = pyframe3dd.ReactionData(
        #    rnode, [moorK[0]], [moorK[1]], [moorK[2]], [moorK[3]], [moorK[4]], [moorK[5]], rigid=RIGID
        # )
        # -----------------------------------

        # ------ frame element data ------------
        element = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n + 1)
        roll = np.zeros(n - 1)

        # Element properties
        Area = inputs["tower_A"]
        Asx = inputs["tower_Asx"]
        Asy = inputs["tower_Asy"]
        J0 = inputs["tower_J0"]
        Ixx = inputs["tower_Ixx"]
        Iyy = inputs["tower_Iyy"]
        E = inputs["tower_E"]
        G = inputs["tower_G"]
        rho = inputs["tower_rho"]

        elements = pyframe3dd.ElementData(element, N1, N2, Area, Asx, Asy, J0, Ixx, Iyy, E, G, roll, rho)
        # -----------------------------------

        # ------ options ------------
        dx = -1.0
        options = pyframe3dd.Options(frame3dd_opt["shear"], frame3dd_opt["geom"], dx)
        # -----------------------------------

        # initialize frame3dd object
        myframe = pyframe3dd.Frame(nodes, reactions, elements, options)

        # ------- enable dynamic analysis ----------
        # Run extra freqs because could get 6 rigid body modes at zero-freq
        lump = 0
        myframe.enableDynamics(
            3 * NFREQ, frame3dd_opt["modal_method"], lump, frame3dd_opt["tol"], frame3dd_opt["shift"]
        )
        # ----------------------------

        # ------ static load case 1 ------------
        # gravity in the X, Y, Z, directions (global)
        gx = 0.0
        gy = 0.0
        gz = -gravity
        load = pyframe3dd.StaticLoadCase(gx, gy, gz)
        myframe.addLoadCase(load)

        # Added mass
        cg_add = np.c_[inputs["platform_total_center_of_mass"], inputs["rna_cg"]]
        add_gravity = False
        mID = np.array([1, n - 1], dtype=np.int_)
        m_fact = inputs["platform_added_mass"].max() / inputs["platform_mass"]
        m_add = np.r_[(1 + m_fact) * inputs["platform_mass"], inputs["rna_mass"]].flatten()
        I_add = np.c_[(1 + m_fact) * inputs["platform_I_total"], inputs["rna_I"]]
        myframe.changeExtraNodeMass(
            mID,
            m_add,
            I_add[0, :],
            I_add[1, :],
            I_add[2, :],
            I_add[3, :],
            I_add[4, :],
            I_add[5, :],
            cg_add[0, :],
            cg_add[1, :],
            cg_add[2, :],
            add_gravity,
        )

        # Debugging
        # myframe.write("floating_tower_debug.3dd")
        # -----------------------------------
        # run the analysis
        try:
            _, _, _, _, _, modal = myframe.run()

            # natural frequncies
            freq = modal.freq
            freq = freq[freq > 1e-1]
            if len(freq) >= NFREQ:
                outputs["f1"] = freq[0]
                outputs["f2"] = freq[1]
                outputs["structural_frequencies"] = freq[:NFREQ]

                # Get all mode shapes in batch
                NFREQ2 = int(NFREQ / 2)
                freq_x, freq_y, freq_z, mshapes_x, mshapes_y, mshapes_z = util.get_xyz_mode_shapes(
                    xyz[:, 2],
                    modal.freq,
                    modal.xdsp,
                    modal.ydsp,
                    modal.zdsp,
                    modal.xmpf,
                    modal.ympf,
                    modal.zmpf,
                    base_slope0=False,
                    rank_and_file=self.options["rank_and_file"],
                )

                outputs["fore_aft_freqs"] = freq_x[:NFREQ2]
                outputs["side_side_freqs"] = freq_y[:NFREQ2]
                outputs["torsion_freqs"] = freq_z[:NFREQ2]
                outputs["fore_aft_modes"] = mshapes_x[:NFREQ2, :]
                outputs["side_side_modes"] = mshapes_y[:NFREQ2, :]
                outputs["torsion_modes"] = mshapes_z[:NFREQ2, :]
        except:
            pass


class FloatingPost(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")
        self.options.declare("n_dlc")

    def setup(self):
        n_dlc = self.options["n_dlc"]

        self.add_input("platform_elem_L", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_J0", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_sigma_y", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_qdyn", NULL * np.ones((NELEM_MAX, n_dlc)), units="Pa")

        # Processed Frame3DD/OpenFAST outputs
        self.add_input("platform_Fz", NULL * np.ones((NELEM_MAX, n_dlc)), units="N")
        self.add_input("platform_Vx", NULL * np.ones((NELEM_MAX, n_dlc)), units="N")
        self.add_input("platform_Vy", NULL * np.ones((NELEM_MAX, n_dlc)), units="N")
        self.add_input("platform_Mxx", NULL * np.ones((NELEM_MAX, n_dlc)), units="N*m")
        self.add_input("platform_Myy", NULL * np.ones((NELEM_MAX, n_dlc)), units="N*m")
        self.add_input("platform_Mzz", NULL * np.ones((NELEM_MAX, n_dlc)), units="N*m")

        self.add_output("constr_platform_stress", NULL * np.ones((NELEM_MAX, n_dlc)))
        self.add_output("constr_platform_shell_buckling", NULL * np.ones((NELEM_MAX, n_dlc)))
        self.add_output("constr_platform_global_buckling", NULL * np.ones((NELEM_MAX, n_dlc)))

    def compute(self, inputs, outputs):
        # Unpack some variables
        opt = self.options["options"]
        n_dlc = self.options["n_dlc"]
        gamma_f = opt["gamma_f"]
        gamma_m = opt["gamma_m"]
        gamma_n = opt["gamma_n"]
        gamma_b = opt["gamma_b"]

        d = inputs["platform_elem_D"]
        nelem = np.where(d == NULL)[0][0]
        d = d[:nelem]
        t = inputs["platform_elem_t"][:nelem]
        h = inputs["platform_elem_L"][:nelem]
        Az = inputs["platform_elem_A"][:nelem]
        Asx = inputs["platform_elem_Asx"][:nelem]
        Jz = inputs["platform_elem_J0"][:nelem]
        Iyy = inputs["platform_elem_Iyy"][:nelem]
        sigy = inputs["platform_elem_sigma_y"][:nelem]
        E = inputs["platform_elem_E"][:nelem]
        G = inputs["platform_elem_G"][:nelem]
        qdyn = inputs["platform_elem_qdyn"][:nelem, :]
        r = 0.5 * d

        # Get loads from Framee3dd/OpenFAST
        Fz = inputs["platform_Fz"][:nelem, :]
        Vx = inputs["platform_Vx"][:nelem, :]
        Vy = inputs["platform_Vy"][:nelem, :]
        Mxx = inputs["platform_Mxx"][:nelem, :]
        Myy = inputs["platform_Myy"][:nelem, :]
        Mzz = inputs["platform_Mzz"][:nelem, :]

        M = np.sqrt(Mxx**2 + Myy**2)
        V = np.sqrt(Vx**2 + Vy**2)

        # Initialize outputs
        outputs["constr_platform_stress"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["constr_platform_shell_buckling"] = NULL * np.ones((NELEM_MAX, n_dlc))
        outputs["constr_platform_global_buckling"] = NULL * np.ones((NELEM_MAX, n_dlc))

        # See http://svn.code.sourceforge.net/p/frame3dd/code/trunk/doc/Frame3DD-manual.html#structuralmodeling
        # print(Fz.shape, Az.shape, M.shape, r.shape, Iyy.shape)
        axial_stress = Fz / Az[:, np.newaxis] + M * (r / Iyy)[:, np.newaxis]
        shear_stress = np.abs(Mzz) / (Jz * r)[:, np.newaxis] + V / Asx[:, np.newaxis]
        hoop_stress = -qdyn * ((r - 0.5 * t) / t)[:, np.newaxis]  # util_con.hoopStress(d, t, qdyn)
        outputs["constr_platform_stress"][:nelem, :] = util_con.vonMisesStressUtilization(
            axial_stress, hoop_stress, shear_stress, gamma_f * gamma_m * gamma_n, sigy.reshape((-1, 1))
        )

        # Use DNV-GL CP202 Method
        check = util_dnvgl.CylinderBuckling(h, d, t, E=E, G=G, sigma_y=sigy, gamma=gamma_f * gamma_b)
        for k in range(n_dlc):
            results = check.run_buckling_checks(
                Fz[k, :], M[k, :], axial_stress[k, :], hoop_stress[k, :], shear_stress[k, :]
            )

            outputs["constr_platform_shell_buckling"][:nelem, k] = results["Shell"]
            outputs["constr_platform_global_buckling"][:nelem, k] = results["Global"]


class FloatingFrame(om.Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]
        nLC = opt["WISDEM"]["n_dlc"]
        n_member = opt["floating"]["members"]["n_members"]
        frame3dd_opt = opt["WISDEM"]["FloatingSE"]["frame3dd"]
        rank_and_file = opt["WISDEM"]["FloatingSE"]["rank_and_file"]

        mem_vars = ["Px", "Py", "Pz", "qdyn"]

        mem_prom = [
            "wind_reference_height",
            "z0",
            "shearExp",
            "cd_usr",
            "cm",
            "rho_air",
            "rho_water",
            "mu_air",
            "mu_water",
            "beta_wind",
            "beta_wave",
            "Uc",
            "Hsig_wave",
            "Tsig_wave",
            "water_depth",
            "yaw",
        ]

        U_prom = []
        for iLC in range(nLC):
            lc = "" if nLC == 1 else str(iLC + 1)
            U_prom.append(f"env{lc}.Uref")

        for k in range(n_member):
            n_full = get_nfull(opt["floating"]["members"]["n_height"][k], nref=2)
            self.add_subsystem(
                f"memload{k}",
                MemberLoads(
                    n_full=n_full,
                    n_lc=nLC,
                    hydro=True,
                    memmax=True,
                ),
                promotes=mem_prom + U_prom + [("joint1", f"member{k}:joint1"), ("joint2", f"member{k}:joint2")],
            )

        self.add_subsystem("loadsys", PlatformLoads(options=opt), promotes=["*"])

        self.add_subsystem("frame", FrameAnalysis(options=opt), promotes=["*"])

        if self.options["modeling_options"]["flags"]["tower"]:
            tow_opt = self.options["modeling_options"]["WISDEM"]["TowerSE"]
            n_height = tow_opt["n_height"]
            n_full_tow = get_nfull(n_height, nref=tow_opt["n_refine"])
            self.add_subsystem(
                "tower",
                TowerModal(n_full=n_full_tow, frame3dd_opt=frame3dd_opt, rank_and_file=rank_and_file),
                promotes=["*"],
            )

        self.add_subsystem("post", FloatingPost(options=opt["WISDEM"]["FloatingSE"], n_dlc=nLC), promotes=["*"])

        for k in range(n_member):
            for var in mem_vars:
                self.connect(f"memload{k}.{var}", f"member{k}:{var}")



import unittest

import numpy as np
import numpy.testing as npt
import wisdem.floatingse.floating_frame as frame
import wisdem.floatingse.floating_system as sys
from wisdem.commonse.cylinder_member import NULL, MEMMAX


class TestFrame(unittest.TestCase):
    def testTetrahedron(self):
        inputs = {}
        outputs = {}

        opt = {}
        opt["floating"] = {}
        opt["flags"] = {}
        opt["flags"]["floating"] = True
        opt["flags"]["tower"] = False
        opt["WISDEM"] = {}
        opt["WISDEM"]["n_dlc"] = 1
        opt["WISDEM"]["FloatingSE"] = {}
        opt["floating"]["members"] = {}
        opt["floating"]["members"]["n_members"] = n_member = 6
        opt["WISDEM"]["FloatingSE"]["frame3dd"] = {}
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["shear"] = True
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["geom"] = True
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["tol"] = 1e-8
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["modal"] = False
        opt["WISDEM"]["FloatingSE"]["gamma_f"] = 1.35  # Safety factor on loads
        opt["WISDEM"]["FloatingSE"]["gamma_m"] = 1.3  # Safety factor on materials
        opt["WISDEM"]["FloatingSE"]["gamma_n"] = 1.0  # Safety factor on consequence of failure
        opt["WISDEM"]["FloatingSE"]["gamma_b"] = 1.1  # Safety factor on buckling
        opt["WISDEM"]["FloatingSE"]["gamma_fatigue"] = 1.755  # Not used
        opt["mooring"] = {}
        opt["mooring"]["n_attach"] = 3

        for k in range(n_member):
            for var in ["Px", "Py", "Pz", "qdyn"]:
                inputs[f"member{k}:{var}"] = NULL * np.ones((MEMMAX, 1))

            inputs[f"member{k}:Px"][:2, :] = 1.0
            inputs[f"member{k}:Py"][:2, :] = 2.0
            inputs[f"member{k}:Pz"][:2, :] = 3.0
            inputs[f"member{k}:qdyn"][:2, :] = 4.0

        myobj = frame.PlatformLoads(options=opt)
        myobj.compute(inputs, outputs)

        npt.assert_equal(outputs["platform_elem_Px1"][6:, :], NULL)
        npt.assert_equal(outputs["platform_elem_Py1"][6:, :], NULL)
        npt.assert_equal(outputs["platform_elem_Pz1"][6:, :], NULL)
        npt.assert_equal(outputs["platform_elem_Px2"][6:, :], NULL)
        npt.assert_equal(outputs["platform_elem_Py2"][6:, :], NULL)
        npt.assert_equal(outputs["platform_elem_Pz2"][6:, :], NULL)
        npt.assert_equal(outputs["platform_elem_qdyn"][6:, :], NULL)

        npt.assert_equal(outputs["platform_elem_Px1"][:6, :], 1.0)
        npt.assert_equal(outputs["platform_elem_Py1"][:6, :], 2.0)
        npt.assert_equal(outputs["platform_elem_Pz1"][:6, :], 3.0)
        npt.assert_equal(outputs["platform_elem_Px2"][:6, :], 1.0)
        npt.assert_equal(outputs["platform_elem_Py2"][:6, :], 2.0)
        npt.assert_equal(outputs["platform_elem_Pz2"][:6, :], 3.0)
        npt.assert_equal(outputs["platform_elem_qdyn"][:6, :], 4.0)

    def testAnalysis(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}

        opt = {}
        opt["floating"] = {}
        opt["WISDEM"] = {}
        opt["WISDEM"]["n_dlc"] = 1
        opt["WISDEM"]["FloatingSE"] = {}
        opt["floating"]["members"] = {}
        opt["floating"]["members"]["n_members"] = n_member = 6
        opt["WISDEM"]["FloatingSE"]["frame3dd"] = {}
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["shear"] = True
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["geom"] = True
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["tol"] = 1e-8
        opt["WISDEM"]["FloatingSE"]["frame3dd"]["modal"] = False
        opt["WISDEM"]["FloatingSE"]["gamma_f"] = 1.35  # Safety factor on loads
        opt["WISDEM"]["FloatingSE"]["gamma_m"] = 1.3  # Safety factor on materials
        opt["WISDEM"]["FloatingSE"]["gamma_n"] = 1.0  # Safety factor on consequence of failure
        opt["WISDEM"]["FloatingSE"]["gamma_b"] = 1.1  # Safety factor on buckling
        opt["WISDEM"]["FloatingSE"]["gamma_fatigue"] = 1.755  # Not used
        opt["mooring"] = {}
        opt["mooring"]["n_attach"] = 3

        inputs["tower_nodes"] = NULL * np.ones((MEMMAX, 3))
        inputs["tower_Rnode"] = NULL * np.ones(MEMMAX)
        for k in range(n_member):
            inputs[f"member{k}:nodes_xyz"] = NULL * np.ones((MEMMAX, 3))
            inputs[f"member{k}:nodes_r"] = NULL * np.ones(MEMMAX)

        for var in [
            "D",
            "t",
            "A",
            "Asx",
            "Asy",
            "Ixx",
            "Iyy",
            "J0",
            "rho",
            "E",
            "G",
            "sigma_y",
            "Px",
            "Py",
            "Pz",
            "qdyn",
        ]:
            inputs["tower_elem_" + var] = NULL * np.ones(MEMMAX)
            for k in range(n_member):
                if var in ["Px", "Py", "Pz", "qdyn"]:
                    inputs[f"member{k}:{var}"] = NULL * np.ones((MEMMAX, 1))
                else:
                    inputs[f"member{k}:section_{var}"] = NULL * np.ones(MEMMAX)

        inputs["member0:nodes_xyz"][:2, :] = np.array([[0, 0, 0], [1, 0, 0]])
        inputs["member1:nodes_xyz"][:2, :] = np.array([[1, 0, 0], [0.5, 1, 0]])
        inputs["member2:nodes_xyz"][:2, :] = np.array([[0.5, 1, 0], [0, 0, 0]])
        inputs["member3:nodes_xyz"][:2, :] = np.array([[0, 0, 0], [0, 0, 1]])
        inputs["member4:nodes_xyz"][:2, :] = np.array([[1, 0, 0], [0, 0, 1]])
        inputs["member5:nodes_xyz"][:2, :] = np.array([[0.5, 1, 0], [0, 0, 1]])
        for k in range(n_member):
            L = np.sqrt(np.sum(np.diff(inputs[f"member{k}:nodes_xyz"][:2, :], axis=0) ** 2))
            inputs[f"member{k}:nodes_r"][:2] = 0.1 * k * np.ones(2)
            inputs[f"member{k}:section_D"][:1] = 2.0
            inputs[f"member{k}:section_t"][:1] = 0.1
            inputs[f"member{k}:section_A"][:1] = 0.5 * k * np.ones(1) + 1
            inputs[f"member{k}:section_Asx"][:1] = 0.5 * k * np.ones(1) + 1
            inputs[f"member{k}:section_Asy"][:1] = 0.5 * k * np.ones(1) + 1
            inputs[f"member{k}:section_Ixx"][:1] = 2 * k * np.ones(1) + 1
            inputs[f"member{k}:section_Iyy"][:1] = 2 * k * np.ones(1) + 1
            inputs[f"member{k}:section_J0"][:1] = 2 * k * np.ones(1) + 1
            inputs[f"member{k}:section_rho"][:1] = 1e3 / (0.5 * k * np.ones(1) + 1) / L
            inputs[f"member{k}:section_E"][:1] = 3 * k * np.ones(1) + 1
            inputs[f"member{k}:section_G"][:1] = 4 * k * np.ones(1) + 1
            inputs[f"member{k}:section_sigma_y"][:1] = 5 * k * np.ones(1) + 1
            inputs[f"member{k}:idx_cb"] = 0
            inputs[f"member{k}:buoyancy_force"] = 1e2
            inputs[f"member{k}:displacement"] = 1e1
            inputs[f"member{k}:center_of_buoyancy"] = inputs[f"member{k}:nodes_xyz"][:2, :].mean(axis=0)
            inputs[f"member{k}:center_of_mass"] = inputs[f"member{k}:nodes_xyz"][:2, :].mean(axis=0)
            inputs[f"member{k}:total_mass"] = 1e3
            inputs[f"member{k}:total_cost"] = 2e3
            inputs[f"member{k}:I_total"] = 1e2 + np.arange(6)
            inputs[f"member{k}:Awater"] = 5.0
            inputs[f"member{k}:Iwater"] = 15.0
            inputs[f"member{k}:added_mass"] = np.arange(6)
            inputs[f"member{k}:ballast_mass"] = 1e2
            inputs[f"member{k}:variable_ballast_capacity"] = 10 + k
            inputs[f"member{k}:variable_ballast_spts"] = np.linspace(0, 0.5, 10)
            inputs[f"member{k}:variable_ballast_Vpts"] = np.arange(10)
            inputs[f"member{k}:waterline_centroid"] = inputs[f"member{k}:nodes_xyz"][:2, :2].mean(axis=0)
            inputs[f"member{k}:Px"][:2, :] = 1.0
            inputs[f"member{k}:Py"][:2, :] = 2.0
            inputs[f"member{k}:Pz"][:2, :] = 3.0
            inputs[f"member{k}:qdyn"][:2, :] = 4.0

        inputs["mooring_neutral_load"] = np.zeros((3, 3))
        inputs["mooring_neutral_load"][:, 0] = [200, -100.0, -100]
        inputs["mooring_neutral_load"][:, 1] = [0.0, 50, -50]
        inputs["mooring_neutral_load"][:, 2] = -1e3
        inputs["mooring_fairlead_joints"] = np.array([[0.0, 0.0, 0.0], [0.5, 1.0, 0.0], [1.0, 0.0, 0.0]])
        inputs["mooring_stiffness"] = 5 * np.eye(6)
        inputs["transition_node"] = inputs["member0:nodes_xyz"][1, :]
        inputs["turbine_mass"] = 1e4
        inputs["turbine_cg"] = np.array([0, 0, 50])
        inputs["turbine_I"] = 1e6 * np.ones(6)
        inputs["transition_piece_mass"] = 1e3
        inputs["transition_piece_cost"] = 3e3
        inputs["rho_water"] = 1e3

        inputs["turbine_F"] = 1e3 * np.ones((3, 1))
        inputs["turbine_M"] = 1e4 * np.ones((3, 1))

        myobj = sys.PlatformFrame(options=opt)
        myobj.node_mem2glob = {}
        myobj.node_glob2mem = {}
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        for k in outputs:
            inputs[k] = outputs[k]
        for k in discrete_outputs:
            discrete_inputs[k] = discrete_outputs[k]
        myobj = sys.PlatformTurbineSystem()
        myobj.compute(inputs, outputs)
        for k in outputs:
            inputs[k] = outputs[k]
        for k in discrete_outputs:
            discrete_inputs[k] = discrete_outputs[k]
        myobj = frame.PlatformLoads(options=opt)
        myobj.compute(inputs, outputs)
        for k in outputs:
            inputs[k] = outputs[k]
        for k in discrete_outputs:
            discrete_inputs[k] = discrete_outputs[k]
        myobj = frame.FrameAnalysis(options=opt)
        myobj.compute(inputs, outputs)
        self.assertTrue(True)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestFrame))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



# tests MoorPy System functionality and results

import numpy as np
import pytest
from numpy.testing import assert_allclose

import wisdem.moorpy as mp

inCBs = [0, 1.0, 10.0]  # friction coefficients as inputs for test_seabed


def test_tensions_swap():
    """Compares two equivalent catenary mooring lines that are defined in opposite directions."""

    ms = mp.System(depth=60)

    # ms.lineTypes['chain'] = getLineProps(120, name='chain')    # add a line type
    ms.setLineType(120, "chain", name="chain")  # add a line type

    ms.addPoint(1, [0, 0, -60])
    ms.addPoint(1, [100, 10, -30])

    # line sloping up from A to B, and another in the opposite order
    ms.addLine(120, "chain", pointA=1, pointB=2)
    ms.addLine(120, "chain", pointA=2, pointB=1)

    ms.initialize()

    # compare tensions
    assert_allclose(
        np.hstack([ms.lineList[0].fA, ms.lineList[0].fB]),
        np.hstack([ms.lineList[1].fB, ms.lineList[1].fA]),
        rtol=0,
        atol=10.0,
        verbose=True,
    )


def test_stiffnesses_swap():
    """Compares two equivalent catenary mooring lines that are defined in opposite directions."""

    ms = mp.System(depth=60)

    # ms.lineTypes['chain'] = getLineProps(120, name='chain')    # add a line type
    ms.setLineType(120, "chain", name="chain")  # add a line type

    ms.addPoint(1, [0, 0, -60])
    ms.addPoint(1, [100, 10, -30])

    # line sloping up from A to B, and another in the opposite order
    ms.addLine(120, "chain", pointA=1, pointB=2)
    ms.addLine(120, "chain", pointA=2, pointB=1)

    ms.initialize()

    # compare stiffnesses
    assert_allclose(
        np.hstack([ms.lineList[0].KA, ms.lineList[0].KB, ms.lineList[0].KAB]),
        np.hstack([ms.lineList[1].KB, ms.lineList[1].KA, ms.lineList[1].KAB]),
        rtol=0,
        atol=10.0,
        verbose=True,
    )


def test_stiffness_body():
    """Tests that the mooring stiffness on a body has the expected relationship to stiffness on points"""

    ms = mp.System(depth=60)

    # ms.lineTypes['chain'] = getLineProps(120, name='chain')    # add a line type
    ms.setLineType(120, "chain", name="chain")  # add a line type

    ms.addPoint(1, [0, 0, -60])
    ms.addPoint(1, [100, 10, -30])

    # line sloping up from A to B, and another in the opposite order
    ms.addLine(120, "chain", pointA=1, pointB=2)
    ms.addLine(120, "chain", pointA=2, pointB=1)

    # create body and attach lines to it
    ms.addBody(1, [0, 0, 0, 0, 0, 0])
    ms.bodyList[0].attachPoint(2, [0, 10, 0])

    ms.initialize()

    # compare stiffnesses
    assert_allclose(
        np.hstack([ms.lineList[0].KA, ms.lineList[0].KB, ms.lineList[0].KAB]),
        np.hstack([ms.lineList[1].KB, ms.lineList[1].KA, ms.lineList[1].KAB]),
        rtol=0,
        atol=10.0,
        verbose=True,
    )


def test_basic():

    depth = 600
    angle = np.arange(3) * np.pi * 2 / 3  # line headings list
    anchorR = 1600  # anchor radius/spacing
    fair_depth = 21
    fairR = 20
    LineLength = 1800
    typeName = "chain"  # identifier string for line type

    # --------------- set up mooring system ---------------------

    # Create blank system object
    ms = mp.System()

    # Set the depth of the system to the depth of the input value
    ms.depth = depth

    # add a line type
    # ms.lineTypes[typeName] = getLineProps(120, name=typeName)
    ms.setLineType(120, typeName, name=typeName)

    # Add a free, body at [0,0,0] to the system (including some properties to make it hydrostatically stiff)
    ms.addBody(0, np.zeros(6), m=1e6, v=1e3, rM=100, AWP=1e6)

    # Set the anchor points of the system
    anchors = []
    for i in range(len(angle)):
        ms.addPoint(1, np.array([anchorR * np.cos(angle[i]), anchorR * np.sin(angle[i]), -ms.depth], dtype=float))
        anchors.append(len(ms.pointList))

    # Set the points that are attached to the body to the system
    bodypts = []
    for i in range(len(angle)):
        ms.addPoint(1, np.array([fairR * np.cos(angle[i]), fairR * np.sin(angle[i]), -fair_depth], dtype=float))
        bodypts.append(len(ms.pointList))
        ms.bodyList[0].attachPoint(
            ms.pointList[bodypts[i] - 1].number, ms.pointList[bodypts[i] - 1].r - ms.bodyList[0].r6[:3]
        )

    # Add and attach lines to go from the anchor points to the body points
    for i in range(len(angle)):
        ms.addLine(LineLength, typeName)
        line = len(ms.lineList)
        ms.pointList[anchors[i] - 1].attachLine(ms.lineList[line - 1].number, 0)
        ms.pointList[bodypts[i] - 1].attachLine(ms.lineList[line - 1].number, 1)

    # ------- simulate it ----------
    ms.initialize()  # make sure everything's connected
    ms.bodyList[0].setPosition([10, 10, 1, 0, 0, 0])  # apply an offset
    ms.solveEquilibrium3()  # equilibrate - see if it goes back to zero!

    # check
    assert_allclose(ms.bodyList[0].r6, np.zeros(6), rtol=0, atol=0.01, verbose=True)


def test_multiseg():
    """Compares a single catenary mooring line with a two-line system."""

    # single line
    ms1 = mp.System()
    ms1.depth = 200
    # ms1.lineTypes['chain'] = getLineProps(120, name='chain')
    ms1.setLineType(120, "chain", name="chain")

    ms1.addPoint(1, [-800, 0, -200])  # anchor point
    ms1.addPoint(1, [0, 0, 0])  # fairlead
    ms1.addLine(860, "chain", pointA=1, pointB=2)
    ms1.initialize()
    ms1.solveEquilibrium3(tol=0.0001)

    # two line
    ms2 = mp.System()
    ms2.depth = 200
    # ms2.lineTypes['chain'] = getLineProps(120, name='chain')
    ms2.setLineType(120, "chain", name="chain")
    ms2.addPoint(1, [-800, 0, -200])  # anchor point
    ms2.addPoint(0, [-205, 0, -100])  # point along line
    ms2.addPoint(1, [0, 0, 0])  # fairlead
    ms2.addLine(630, "chain", pointA=1, pointB=2)
    ms2.addLine(230, "chain", pointA=2, pointB=3)
    ms2.initialize()
    ms2.solveEquilibrium3(tol=0.0001)

    # compare tensions
    assert_allclose(
        np.hstack([ms1.pointList[0].getForces(), ms1.pointList[-1].getForces()]),
        np.hstack([ms2.pointList[0].getForces(), ms2.pointList[-1].getForces()]),
        rtol=0,
        atol=10.0,
        verbose=True,
    )


def test_multiseg_seabed():
    """Compares a single catenary mooring line with a three-line system where two of the segments are fully along the seabed."""

    # single line
    ms1 = mp.System()
    ms1.depth = 200
    # ms1.lineTypes['chain'] = getLineProps(120, name='chain')
    ms1.setLineType(120, "chain", name="chain")
    ms1.addPoint(1, [-800, 0, -200])  # anchor point
    ms1.addPoint(1, [0, 0, 0])  # fairlead
    ms1.addLine(860, "chain", pointA=1, pointB=2)
    ms1.initialize()
    ms1.solveEquilibrium3(tol=0.0001)

    # three line
    ms2 = mp.System()
    ms2.depth = 200
    # ms2.lineTypes['chain'] = getLineProps(120, name='chain')
    ms2.setLineType(120, "chain", name="chain")
    ms2.addPoint(1, [-800, 0, -200])  # anchor point
    ms2.addPoint(0, [-700, 0, -200])  # point along line
    ms2.addPoint(0, [-600, 0, -200])  # point along line
    ms2.addPoint(1, [0, 0, 0])  # fairlead
    ms2.addLine(100, "chain", pointA=1, pointB=2)
    ms2.addLine(100, "chain", pointA=2, pointB=3)
    ms2.addLine(660, "chain", pointA=3, pointB=4)
    ms2.initialize()
    ms2.solveEquilibrium3(tol=0.0001)

    # compare tensions
    assert_allclose(
        np.hstack([ms1.pointList[0].getForces(), ms1.pointList[-1].getForces()]),
        np.hstack([ms2.pointList[0].getForces(), ms2.pointList[-1].getForces()]),
        rtol=0,
        atol=10.0,
        verbose=True,
    )


def test_basicU():
    """Compares a system with a U-shape line with seabed contact with an equivalent case
    that has a node along the U."""

    # a seabed contact case with 2 lines to form a U
    ms1 = mp.System()

    ms1.depth = 100

    # ms1.lineTypes['chain'] = getLineProps(120, name='chain')
    ms1.setLineType(120, "chain", name="chain")

    ms1.addPoint(1, [-200, 0, -100])  # anchor point
    ms1.addPoint(0, [-100, 0, -50], m=0, v=50)  # float
    ms1.addPoint(0, [0, 0, -100])  # midpoint
    ms1.addPoint(0, [100, 0, -40], m=0, v=50)  # float
    ms1.addPoint(1, [200, 0, -100])  # anchor point

    ms1.addLine(120, "chain", pointA=1, pointB=2)
    ms1.addLine(125, "chain", pointA=2, pointB=3)
    ms1.addLine(125, "chain", pointA=3, pointB=4)
    ms1.addLine(120, "chain", pointA=4, pointB=5)

    # a seabed contact case with single U line
    msU = mp.System()

    msU.depth = 100

    # msU.lineTypes['chain'] = getLineProps(120, name='chain')
    msU.setLineType(120, "chain", name="chain")

    msU.addPoint(1, [-200, 0, -100])  # anchor point
    msU.addPoint(0, [-100, 0, -50], m=0, v=50)  # float
    msU.addPoint(0, [100, 0, -40], m=0, v=50)  # float
    msU.addPoint(1, [200, 0, -100])  # anchor point

    msU.addLine(120, "chain", pointA=1, pointB=2)
    msU.addLine(250, "chain", pointA=2, pointB=3)
    msU.addLine(120, "chain", pointA=3, pointB=4)

    # ------- simulate it ----------
    ms1.initialize()  # make sure everything's connected
    msU.initialize()  # make sure everything's connected

    ms1.solveEquilibrium3(tol=0.0001)  # equilibrate - see if it goes back to zero!
    msU.solveEquilibrium3(tol=0.0001)  # equilibrate - see if it goes back to zero!

    # compare floating point positions
    assert_allclose(
        np.hstack([ms1.pointList[1].r, ms1.pointList[3].r]),
        np.hstack([msU.pointList[1].r, msU.pointList[2].r]),
        rtol=0,
        atol=0.001,
        verbose=True,
    )


@pytest.mark.parametrize("CB", inCBs)
def test_seabed(CB):
    """Compares a single catenary mooring line along the seabed with a two-line system
    where the point is on the seabed, with different friction settings."""

    # single line
    ms1 = mp.System()
    ms1.depth = 200
    # ms1.lineTypes['chain'] = getLineProps(120, name='chain')
    ms1.setLineType(120, "chain", name="chain")
    ms1.addPoint(1, [-800, 0, -200])  # anchor point
    ms1.addPoint(1, [0, 0, 0])  # fairlead
    ms1.addLine(860, "chain", pointA=1, pointB=2, cb=CB)
    ms1.initialize()
    ms1.solveEquilibrium3(tol=0.0001)

    # two line
    ms2 = mp.System()
    ms2.depth = 200
    # ms2.lineTypes['chain'] = getLineProps(120, name='chain')
    ms2.setLineType(120, "chain", name="chain")
    ms2.addPoint(1, [-800, 0, -200])  # anchor point
    ms2.addPoint(0, [-405, 0, -150])  # midpoint
    ms2.addPoint(1, [0, 0, 0])  # fairlead
    ms2.addLine(430, "chain", pointA=1, pointB=2, cb=CB)
    ms2.addLine(430, "chain", pointA=2, pointB=3, cb=CB)
    ms2.initialize()
    ms2.solveEquilibrium3(tol=0.0001)

    # compare tensions
    assert_allclose(
        np.hstack([ms1.pointList[0].getForces(), ms1.pointList[-1].getForces()]),
        np.hstack([ms2.pointList[0].getForces(), ms2.pointList[-1].getForces()]),
        rtol=0,
        atol=10.0,
        verbose=True,
    )


if __name__ == "__main__":

    # test_basic()

    """
      # a seabed contact case with 2 lines to form a U
    ms1 = mp.System()

    ms1.depth = 100

    ms1.lineTypes['chain'] = getLineProps(120, name='chain')

    ms1.addPoint(1, [-200, 0 , -100])                 # anchor point
    ms1.addPoint(1, [-100, 0 ,  -50], m=0, v=50)      # float
    ms1.addPoint(0, [   0, 0 , -100])                 # midpoint
    ms1.addPoint(1, [ 100, 0 ,  -40], m=0, v=50)      # float
    ms1.addPoint(1, [ 200, 0 , -100])                 # anchor point

    ms1.addLine(120, 'chain', pointA=1, pointB=2)
    ms1.addLine(125, 'chain', pointA=2, pointB=3)
    ms1.addLine(125, 'chain', pointA=3, pointB=4)
    ms1.addLine(120, 'chain', pointA=4, pointB=5)


    # a seabed contact case with single U line
    ms = mp.System()

    ms.depth = 100

    ms.lineTypes['chain'] = getLineProps(120, name='chain')

    ms.addPoint(1, [-200, 0 , -100])                 # anchor point
    ms.addPoint(0, [-100, 0 ,  -50], m=0, v=50)      # float
    ms.addPoint(0, [ 100, 0 ,  -40], m=0, v=50)      # float
    ms.addPoint(1, [ 200, 0 , -100])                 # anchor point

    ms.addLine(120, 'chain', pointA=1, pointB=2)
    ms.addLine(250, 'chain', pointA=2, pointB=3)
    ms.addLine(120, 'chain', pointA=3, pointB=4)


    # ------- simulate it ----------
    ms1.initialize()                                             # make sure everything's connected
    ms.initialize()                                             # make sure everything's connected

    fig, ax = ms1.plot(color='b')
    ms.plot(ax=ax, color='k')

    #ms.display=2

    ms1.solveEquilibrium3(maxIter=20)                                      # equilibrate - see if it goes back to zero!
    #ms.solveEquilibrium3(maxIter=20)                                      # equilibrate - see if it goes back to zero!

    ms1.plot(ax=ax, color='g')
    #ms.plot(ax=ax, color='r')

    # compare
    print(ms1.pointList[1].getForces())
    print(ms.pointList[1].getForces())
    print(ms1.pointList[3].getForces())
    print(ms.pointList[2].getForces())

    print(ms1.pointList[1].getStiffnessA())
    print(ms.pointList[1].getStiffnessA())
    print(ms1.pointList[3].getStiffnessA())
    print(ms.pointList[2].getStiffnessA())
    """

    """
    # a seabed contact case with single U line
    ms = mp.System()
    ms.depth = 100
    ms.lineTypes['chain'] = getLineProps(120, name='chain')
    # a suspension bridge shape
    ms.addPoint(1, [-200, 0 , -100])                 # 1 anchor point
    ms.addPoint(0, [-100, 0 ,  -50], m=0, v=50)      # 2 float
    ms.addPoint(0, [ 100, 0 ,  -40], m=0, v=90)      # 3 float
    ms.addPoint(1, [ 200, 0 , -100])                 # 4 anchor point
    ms.addLine(120, 'chain', pointA=1, pointB=2)
    ms.addLine(250, 'chain', pointA=2, pointB=3)
    ms.addLine(120, 'chain', pointA=3, pointB=4)
    # something in a new direction
    ms.addPoint(1, [ 100,  200, -100])               # 5 north anchor
    ms.addPoint(0, [ 100,  100,  -50])               # 6 point along seabed
    ms.addPoint(1, [ 100, -100,  -50])               # 7 south anchor in midair
    ms.addLine(120, 'chain', pointA=5, pointB=6)
    ms.addLine(120, 'chain', pointA=6, pointB=3)
    ms.addLine(250, 'chain', pointA=7, pointB=3)
    # ------- simulate it ----------
    ms.initialize()
    fig,ax = ms.plot(color='g')
    ms.solveEquilibrium3(tol=0.0001)
    ms.plot(color=[1,0,0,0.5], ax=ax)

    #>>> add a test for stiffnesses at each end of lines that have seaed contact, comparing U shape to two catenary lines. Also look at system-level stiffnes.

    plt.show()
    """
